# Scraper and Back Scraper for New York Commercial Division
# CourtID: nysupct
# Court Short Name: NY
# Author: Andrei Chelaru
# Reviewer:
# Date: 2015-10-30

from juriscraper.opinions.united_states.state import nyappterm_1st


class Site(nyappterm_1st.Site):

    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court = 'Commercial+Division'
        self.parameters.update({'court': self.court})


from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'https://supremecourt.nebraska.gov/courts/supreme-court/opinions'
        self.status = 'Published'

    def _process_html(self):
        path = {
            'group': '//div[contains(@class, "view-grouping")]',
            'date': './/span[contains(@class, "date-display-single")]',
            'row': './/tbody/tr',
        }
        for group in self.html.xpath(path['group']):
            date_element = group.xpath(path['date'])
            if not date_element:
                continue
            date = date_element[0].text_content()
            for row in group.xpath(path['row']):
                url = row.xpath('.//td[3]//a/@href')

                # skip rows without url in third cell
                if not url:
                    continue

                self.cases.append({
                    'date': date,
                    'docket': row.xpath('.//td[1]')[0].text_content(),
                    'name': row.xpath('.//td[3]')[0].text_content(),
                    'neutral_citation': row.xpath('.//td[2]')[0].text_content(),
                    'url': url[0],
                })



"""Scraper for [Full name of court]
CourtID: [unique abbreviation to be used by software/filesystem]
Court Short Name: [standard abbreviation used in citations]
Author:
Reviewer:
History:
  YYYY-MM-DD: Created by XXX
"""


from lxml import html
from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import titlecase
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://court-url.gov/some-path.html'
        # Complete this variable if you create a backscraper.
        self.back_scrape_iterable = None

        # if a POST, use these two attributes, otherwise, delete them.
        self.parameters = {}
        self.method = 'POST'
        self.uses_selenium = False

    """
    Required fields - InsanityExceptions will be thrown if omitted.

    Remove this comment before submission.
    """
    def _get_download_urls(self):
        """
        This is an example of a basic piece of meta data accessible with a
        simple XPath query.

        On a good site, most of your methods will follow this pattern.

        Note that relative URLs are made absolute by the AbstractSite
        object's cleanup routines, so doing so here is not needed.
        """
        path = '//path/to/text/text()'
        return list(self.html.xpath(path))

    def _get_case_names(self):
        """
        This example demonstrates how to extract text from an element that
        may contain other elements.

        For example, this will work well on something like:
           <strong>Nadim v. <em>Jenny</em></strong>

        Resulting in text like:
           Nadim v. Jenny

        Note that titlecase() should be used here in the case that the case
        names are provided in uppercase. Use the titlecase function on
        cases where the name is provided in uppercase only.
        """
        case_names = []
        for e in self.html.xpath('//path/to/an/element/p'):
            s = html.tostring(e, method='text', encoding='unicode')
            case_names.append(titlecase(s))
        return case_names

    def _get_case_dates(self):
        """
        This is an example of a date field. Note that the format string
        will likely need to be updated to  match the date formats
        on the site you are scraping. The datetime formats can be found
        here: http://docs.python.org/2/library/datetime.html
        """
        path = '//path/to/text/text()'
        return [convert_date_string(date_string) for date_string in
                self.html.xpath(path)]

    def _get_precedential_statuses(self):
        """
        In most cases, this field should be normalized to either
        'Published' or 'Unpublished', as below.
        """
        statuses = []
        for e in self.html.xpath('//path/to/text/text()'):
            s = html.tostring(e, method='text', encoding='unicode')
            if 'Opinion' in s:
                statuses.append('Published')
            elif 'Nonprecedential' in s:
                statuses.append('Unpublished')
            else:
                statuses.append('Unknown')
        return statuses

    """
      High priority fields

      Remove this comment and any unused methods before submission
    """
    def _get_docket_numbers(self):
        """
        This is typically of the form ##-####
        """
        return None

    def _get_neutral_citations(self):
        """
        This is often of the form year, state abbreviation, sequential number
        as in '2013 Neb. 12' which would be the 12th opinion issued in 2013.
        """
        return None

    def _get_judges(self):
        """
        For the name of the judge authoring the lead opinion
        """
        return None

    def _get_lower_courts(self):
        """
        If the document is an opinion of an appellate court, then the court
        from which the case was appealed would go here. So, a 9th Circuit
        case might be appealed from the 'N.D. Cal.' which would go here.
        """
        return None

    def _get_nature_of_suit(self):
        """
        Some courts provide a NOS code or phrase giving the subject matter
        of the case, such as 'Contract' 'Civil Rights' 'Immigration' etc.
        """
        return None

    def _get_summaries(self):
        """
        Some courts provide a summary of the case or the opinion's holding.
        """
        return None

    def _get_west_citations(self):
        """
        You are unlikely to encounter these unless it's a backscraper because
        these citations only exist once the document is in a bound volume.
        """
        return None

    def _get_west_state_citations(self):
        """
        You are unlikely to encounter these unless it's a backscraper because
        these citations only exist once the document is in a bound volume.
        """
        return None

    """
    Optional fields

    Remove this comment and any unused methods before submission
    """
    def _get_adversary_numbers(self):
        """
        Similar to a docket number, but found only in bankruptcy cases.
        """
        return None

    def _get_causes(self):
        """
        Some courts will give the plaintiff's or appellant's cause of action
        such as 'Employment' 'Trademark' etc.
        """
        return None

    def _get_dispositions(self):
        """
        Affirmed, Reversed, Vacated; if longer than a sentence, should
        probably use _get_summaries instead.
        """
        return None

    def _get_docket_attachment_numbers(self):
        """
        A document appears on a docket. Each line on the docket has a number.
        Item number 12, for example, might be a declaration with several
        attachments, numbered 1-5. That number, from 1-5, goes here.
        """
        return None

    def _get_docket_document_numbers(self):
        """
        A document appears on a docket. Each line on the docket has a number.
        Those numbers go here.
        """
        return None

    def _get_lower_court_judges(self):
        """
        If the document is an opinion of an appellate court, then the court
        from which the case was appealed is the lower court. So, a 9th Cir.
        case might be appealed from the N.D. Cal. and the judge from that
        court that handled the case would go here.
        """
        return None

    def _get_lower_court_numbers(self):
        """
        If the document is an opinion of an appellate court, then the court
        from which the case was appealed is the lower court. So, a 9th Cir.
        case might be appealed from the N.D. Cal. and the docket number or
        other relevant case numbers from that court would go here.
        """
        return None

    """
    Optional methods for special purposes
    """
    @staticmethod
    def cleanup_content(content):
        """
        Given the HTML from a page, this method is used to strip it down to
        its bare essentials.

        Jurisdictions post their content in an HTML page where the headers,
        foots and other content must be stripped after the page has been
        downloaded by the caller. The intention of this method is that the
        caller is able to call the Site object and do all the meta data
        parsing necessary. Once that's complete, the caller can call this
        method, stripping the html content down to its essentials, before
        saving.
        """
        return content

    def _download_backwards(self, date_str):
        """
        This is a simple method that can be used to generate Site objects
        that can be used to paginate through a court's entire website.

        This method is usually called by a backscraper caller (see the
        one in CourtListener/alert/scrapers for details), and typically
        modifies aspects of the Site object's attributes such as Site.url.

        A simple example has been provided below. The idea is that the
        caller runs this method with a different variable on each iteration.
        That variable is often a date that is getting iterated or is simply
        a index (i), that we iterate upon.

        This can also be used to hold notes useful to future backscraper
        development.
        """
        self.url = 'http://example.com/new/url/%s' % date_str
        self.html = self._download()



# Author: Phil Ardery
# Contact: https://www.ndcourts.gov/contact-us
# Date created: 2019-02-28

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.exceptions import InsanityException
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'https://www.ndcourts.gov/supreme-court/recent-opinions?pageSize=100'
        self.cases = []

    def _process_html(self):
        for row in self.html.xpath('//table//div[@class="row"]'):
            case = self.case_fields_extract(row)
            self.case_fields_validate(case)
            case = self.case_fields_sanitize(case)
            self.cases.append(case)

    def case_fields_extract(self, row):
        text_lines = row.xpath('./div/p[1]/text()')
        text_lines = [l.strip() for l in text_lines if l.strip()]  # Remove empty lines
        line_count = len(text_lines)
        return {
            'citation': text_lines[0],
            'docket': text_lines[1],
            'date': text_lines[2],
            'name': row.xpath('.//a')[0].text_content().strip(),
            'nature': text_lines[3],
            'judge': text_lines[4],
            'summary': ' '.join(text_lines[5:line_count]) if line_count > 5 else '',
            'url': row.xpath('.//button/@onclick')[0].split("'")[1],
        }

    def case_fields_validate(self, case):
        if 'ND' not in case['citation']:
            raise InsanityException('Invalid citation: %s' % case['citation'])
        if not case['docket'].startswith('Docket No.:'):
            raise InsanityException('Invalid docket raw string: %s' % case['docket'])
        if not case['date'].startswith('Filing Date:'):
            raise InsanityException('Invalid date string raw string: %s' % case['date'])
        if not case['nature'].startswith('Case Type:'):
            raise InsanityException('Invalid type raw string: %s' % case['nature'])
        if not case['judge'].startswith('Author:'):
            raise InsanityException('Invalid author raw string: %s' % case['judge'])

    def case_fields_sanitize(self, case):
        for field in ['date', 'docket', 'judge', 'nature']:
            case[field] = case[field].split(':', 1)[1].strip()
        return case

    def _get_download_urls(self):
        return [case['url'] for case in self.cases]

    def _get_case_names(self):
        return [case['name'] for case in self.cases]

    def _get_case_dates(self):
        return [convert_date_string(case['date']) for case in self.cases]

    def _get_docket_numbers(self):
        return [case['docket'] for case in self.cases]

    def _get_nature_of_suit(self):
        return [case['nature'] for case in self.cases]

    def _get_neutral_citations(self):
        return [case['citation'] for case in self.cases]

    def _get_judges(self):
        return [case['judge'] for case in self.cases]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.cases)

    def _get_summaries(self):
        return [case['summary'] for case in self.cases]



"""Scraper for Army Court of Criminal Appeals
CourtID: acca
Reviewer: None
History:
  2015-01-08: Created by mlr
  2016-03-17: Website appears to be dead. Scraper disabled in __init__.py.
  2019-09-11: Rewritten by arderyp to user Linear pattern
"""

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'https://www.jagcnet.army.mil/85257546006DF36B/ODD?OpenView&Count=-1'
        self.status = 'Published'

    def _process_html(self):
        for row in self.html.xpath('//tr[@class="domino-viewentry"]'):
            cell_three = row.xpath('.//td[3]')[0]
            text = cell_three.text_content()
            text_parts = text.split('-', 1)
            hrefs = cell_three.xpath('.//a/@href')

            # skip rows without link to opinion
            if not hrefs:
                continue

            self.cases.append({
                'date': row.xpath('.//td[1]')[0].text_content(),
                'docket': text_parts[0],
                'name': text_parts[1],
                'url': hrefs[0],
            })



from juriscraper.OpinionSiteLinear import OpinionSiteLinear
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://nvcourts.gov/Supreme/Decisions/Advance_Opinions/'
        self.status = 'Published'

    def _process_html(self):
        table_row_path = '//table[@style="border-spacing:0px"]//tr'
        for row in self.html.xpath(table_row_path):
            citation = row.xpath('.//td[1]')[0].text_content().strip()
            docket = row.xpath('.//td[2]')[0].text_content()
            name = row.xpath('.//td[3]')[0].text_content()
            url = row.xpath('.//td[4]//a/@href')
            date_string = row.xpath('.//td[4]')[0].text_content()
            year = convert_date_string(date_string).year

            self.cases.append({
                'date': date_string,
                'docket': docket,
                'name': name if name else "Unknown",
                'neutral_citation': '%d NV %s' % (year, citation),
                'url': url[0],

            })



# Author: Michael Lissner
# Date created: 2013-06-11

import re
from datetime import datetime
from lxml import html
from selenium import webdriver
from time import sleep
from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import titlecase


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.sdjudicial.com/sc/scopinions.aspx'
        self.uses_selenium = True

    def _get_download_urls(self):
        path = '//tr[contains(@id, "ctl00xmainCopyxWGOpinions_r")]//a[2]/@href'
        return list(self.html.xpath(path))

    def _get_case_names(self):
        if self.year == 2005:
            strings = []
            path = '//tr[contains(@id, "ctl00xmainCopyxWGOpinions_r")]/td[2]'
            for e in self.html.xpath(path):
                strings.append(html.tostring(e, method='text', encoding='unicode'))
        else:
            path = '//tr[contains(@id, "ctl00xmainCopyxWGOpinions_r")]/td[2]/text()'
            strings = list(self.html.xpath(path))
        case_names = []
        for s in strings:
            try:
                case_name = re.search(r'(.*)(\d{4} S\.?D\.? \d{1,4})', s, re.MULTILINE).group(1)
                case_names.append(titlecase(case_name.lower()))
            except AttributeError:
                print("AttributeError on: %s" % titlecase(s))
                if 'myrl' in s.lower() and self.year == 2000:
                    case_names.append('Lends His Horse v. Myrl & Roy')
                elif 'springer' in s.lower() and self.year == 2000:
                    case_names.append('State v. Springer-Ertl')
                elif 'formatting provided courtesy' in s.lower() and self.year == 2000:
                    case_names.append('Lois F. Henry v. Harold L. Henry')
                elif 'spec. v. avera' in s.lower() and self.year == 2001:
                    case_names.append('Drs., Residents, and Orth. Surg. Spec. v. Avera St. Luke')
                elif 'clausen' in s.lower() and self.year == 2003:
                    case_names.append('Kelly Clausen v. Northern Plains Recycling, Fireman')
                elif 'burkhart' in s.lower() and self.year == 2003:
                    case_names.append('Burkhart v. Lillehaug and Lillihaug')
                elif 'bennett' in s.lower() and self.year == 2003:
                    case_names.append('State of South Dakota, ex rel., Megan Bennett v. Thomas G. Peterson')
                elif 'duane' in s.lower() and self.year == 2004:
                    case_names.append('State of South Dakota v. Duane J. St. John')
                elif 'square' in s.lower() and self.year == 2005:
                    case_names.append('Town Square Limited Partnership v. Clay County Board of Equalization')
                elif 'toft' in s.lower() and self.year == 2006:
                    case_names.append('Toft v. Toft & Stratmeyer')
                elif 'masteller' in s.lower() and self.year == 2006:
                    case_names.append('Masteller v. Champion Home')
                elif 'nuzum' in s.lower() and self.year == 2006:
                    case_names.append('State v. Nuzum')
                elif 'aaberg' in s.lower() and self.year == 2006:
                    case_names.append('State v. Aaberg')
                elif 'bryan' in s.lower() and self.year == 2006:
                    case_names.append("O'Bryan v. Ashland")
                elif 'j.d.m.c.' in s.lower() and self.year == 2007:
                    case_names.append('In Re the Matter of J.D.M.C.')
                elif 'argus' in s.lower() and self.year == 2007:
                    case_names.append('Argus Leader v. Hagen')
                elif 'nedved' in s.lower() and self.year == 2007:
                    case_names.append('Countryside S. Homeowners Assoc. v. Nedved')
                elif 'gisi' in s.lower() and self.year == 2007:
                    case_names.append('Gisi v. Gisi')
                elif 'western' in s.lower() and self.year == 2007:
                    case_names.append('Great Western Bank v. H&E Enterprises')
                elif 'canyon' in s.lower() and self.year == 2007:
                    case_names.append('Jackson v. Canyon Place Homeowners')
                elif 'vansteenwyk' in s.lower() and self.year == 2007:
                    case_names.append('Vanteenwyk v. Baumgartner Trees')
                elif 'jewett' in s.lower() and self.year == 2011:
                    case_names.append('Jewett v. Real Tuff')
                elif 'jensen' in s.lower() and self.year == 2011:
                    case_names.append('State v. Jensen')
                elif 'cook' in s.lower() and self.year == 2011:
                    case_names.append('Orr v. Cook')
                elif 'living' in s.lower() and self.year == 2011:
                    case_names.append('Benson Living Trust')
                else:
                    raise AttributeError
        return case_names

    def _get_case_dates(self):
        path = '//tr[contains(@id, "ctl00xmainCopyxWGOpinions_r")]/td[1]/@uv'
        return [datetime.strptime(date_string, '%m/%d/%Y').date()
                for date_string in self.html.xpath(path)]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_names)

    def _get_docket_numbers(self):
        path = '//tr[contains(@id, "ctl00xmainCopyxWGOpinions_r")]//a[2]/@href'
        docket_numbers = []
        for url in self.html.xpath(path):
            try:
                # New style: easy
                docket_numbers.append(re.search(r'(\d+).pdf', url).group(1))
            except AttributeError:
                docket_numbers.append(None)
        return docket_numbers

    def _get_neutral_citations(self):
        if self.year == 2005:
            strings = []
            path = '//tr[contains(@id, "ctl00xmainCopyxWGOpinions_r")]/td[2]'
            for e in self.html.xpath(path):
                strings.append(html.tostring(e, method='text', encoding='unicode'))
        else:
            path = '//tr[contains(@id, "ctl00xmainCopyxWGOpinions_r")]/td[2]/text()'
            strings = list(self.html.xpath(path))
        neutral_cites = []
        for s in strings:
            try:
                neutral_cite = re.search(r'(.*)(\d{4} S\.?D\.? \d{1,4})', s, re.MULTILINE).group(2)
                neutral_cites.append(titlecase(neutral_cite))
            except AttributeError:
                if 'myrl' in s.lower() and self.year == 2000:
                    neutral_cites.append('2000 SD 146')
                elif 'springer' in s.lower() and self.year == 2000:
                    neutral_cites.append('2000 SD 56')
                elif 'formatting provided courtesy' in s.lower() and self.year == 2000:
                    neutral_cites.append('2000 SD 4')
                elif 'spec. v. avera' in s.lower() and self.year == 2001:
                    neutral_cites.append('2001 SD 9')
                elif 'clausen' in s.lower() and self.year == 2003:
                    neutral_cites.append('2003 SD 63')
                elif 'burkhart' in s.lower() and self.year == 2003:
                    neutral_cites.append('2003 SD 62')
                elif 'bennett' in s.lower() and self.year == 2003:
                    neutral_cites.append('2003 SD 16')
                elif 'duane' in s.lower() and self.year == 2004:
                    neutral_cites.append('2004 SD 15')
                elif 'square' in s.lower() and self.year == 2005:
                    neutral_cites.append('2005 SD 99')
                elif 'toft' in s.lower() and self.year == 2006:
                    neutral_cites.append('2006 SD 91')
                elif 'masteller' in s.lower() and self.year == 2006:
                    neutral_cites.append('2006 SD 90')
                elif 'nuzum' in s.lower() and self.year == 2006:
                    neutral_cites.append('2006 SD 89')
                elif 'aaberg' in s.lower() and self.year == 2006:
                    neutral_cites.append('2006 SD 58')
                elif 'bryan' in s.lower() and self.year == 2006:
                    neutral_cites.append('2006 SD 56')
                elif 'j.d.m.c.' in s.lower() and self.year == 2007:
                    neutral_cites.append('2007 SD 97')
                elif 'argus' in s.lower() and self.year == 2007:
                    neutral_cites.append('2007 SD 96')
                elif 'nedved' in s.lower() and self.year == 2007:
                    neutral_cites.append('2007 SD 70')
                elif 'gisi' in s.lower() and self.year == 2007:
                    neutral_cites.append('2007 SD 39')
                elif 'western' in s.lower() and self.year == 2007:
                    neutral_cites.append('2007 SD 38')
                elif 'canyon' in s.lower() and self.year == 2007:
                    neutral_cites.append('2007 SD 37')
                elif 'vansteenwyk' in s.lower() and self.year == 2007:
                    neutral_cites.append('2007 SD 36')
                elif 'jewett' in s.lower() and self.year == 2011:
                    neutral_cites.append('2011 S.D.33')
                elif 'jensen' in s.lower() and self.year == 2011:
                    neutral_cites.append('2011 S.D.32')
                elif 'cook' in s.lower() and self.year == 2011:
                    neutral_cites.append('2011 S.D.31')
                elif 'living' in s.lower() and self.year == 2011:
                    neutral_cites.append('2011 S.D.30')
                else:
                    raise AttributeError
        return neutral_cites

    def _download_backwards(self, year):
        self.year = year
        browser = webdriver.PhantomJS()
        browser.get(self.url)
        elems = browser.find_elements_by_class_name('igeb_ItemLabel')
        elem = [elem for elem in elems if elem.text == str(year)][0]
        elem.click()
        sleep(5)

        text = browser.page_source
        html_tree = html.fromstring(text)
        html_tree.make_links_absolute(self.url)

        remove_anchors = lambda url: url.split('#')[0]
        html_tree.rewrite_links(remove_anchors)

        self.html = html_tree
        self.status = 200
        browser.quit()



"""Scraper for Colorado Appeals Court
CourtID: coloctapp
Court Short Name: Colo. Ct. App.
"""

from juriscraper.opinions.united_states.state import colo


class Site(colo.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "http://www.cobar.org/For-Members/Opinions-Rules-Statutes/Colorado-Court-of-Appeals-Opinions"
        self.base_path = "//div[@id='dnn_ctr2514_ModuleContent']/ul/li/a"
        self.next_subpage_path = "//a[@id='dnn_ctr2517_DNNArticle_List_MyArticleList_MyPageNav_cmdNext']"



from juriscraper.lib.html_utils import get_row_column_links
from juriscraper.lib.html_utils import get_row_column_text
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'https://appellate-records.courts.alaska.gov/CMSPublic/Home/Opinions?isCOA=False'
        self.status = 'Published'

    def _process_html(self):
        for table in self.html.xpath('//table'):
            date = table.xpath('./preceding-sibling::h5')[0].text_content()
            for row in table.xpath('.//tr'):
                if row.text_content().strip():
                    
                    # skip rows without PDF links in first column
                    try:
                        url = get_row_column_links(row, 1)
                    except IndexError:
                        continue

                    self.cases.append({
                        'date': date,
                        'docket': get_row_column_text(row, 3),
                        'name': get_row_column_text(row, 4),
                        'neutral_citation': get_row_column_text(row, 2),
                        'url': url,
                    })



# Contact: webmaster@illinoiscourts.gov, 217-558-4490, 312-793-3250
# History:
#   2013-08-16: Created by Krist Jin
#   2014-12-02: Updated by Mike Lissner to remove the summaries code.
#   2016-02-26: Updated by arderyp: simplified thanks to new id attribute identifying decisions table
#   2016-03-27: Updated by arderyp: fixed to handled non-standard formatting

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.state.il.us/court/Opinions/recent_supreme.asp'
        # Sometimes court forgets pdf link in 5th cell.
        # So, only process rows with a link in 5th cell.
        limiter = 'contains(./td[5]//a[1]/@href, "pdf")'
        self.cell_path = '//table[@id="decisions"]//tr[' + limiter + ']/td[%d]'

    def _get_download_urls(self):
        path = '%s//a[1]/@href' % (self.cell_path % 5)
        return [href for href in self.html.xpath(path)]

    def _get_case_names(self):
        path = '%s//a[1]' % (self.cell_path % 5)
        return [anchor.text_content().strip() for anchor in self.html.xpath(path)]

    def _get_case_dates(self):
        path = self.cell_path % 1
        return [convert_date_string(cell.text_content()) for cell in self.html.xpath(path)]

    def _get_precedential_statuses(self):
        statuses = []
        path = '%s//div' % (self.cell_path % 3)
        for div in self.html.xpath(path):
            text = div.xpath('strong/text()')
            if text and 'NRel' in text:
                statuses.append('Unpublished')
            else:
                statuses.append('Published')
        return statuses

    def _get_docket_numbers(self):
        docket_numbers = []
        path = '%s//div' % (self.cell_path % 3)
        for div in self.html.xpath(path):
            text = div.xpath('text()')
            text = ''.join(text).replace('cons.', '')
            docket_numbers.append(' '.join(text.split()))
        return docket_numbers

    def _get_neutral_citations(self):
        path = self.cell_path % 4
        return [cell.text_content().strip() for cell in self.html.xpath(path)]



# Scraper for California's Superior Court Appellate Division
# CourtID: calctapp_app_div
# Court Short Name: Cal. Ct. App.

from juriscraper.opinions.united_states.state import calctapp_1st


class Site(calctapp_1st.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court_code = 'J'
        self.division = 'App. Div.'



from juriscraper.opinions.united_states.federal_appellate import scotus_chambers


class Site(scotus_chambers.Site):

    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.precedential = 'Relating-to'
        self.court = 'relatingtoorders'



# Scraper for Florida 2nd District Court of Appeal Written
# CourtID: flaapp2
# Court Short Name: flaapp2

from . import fladistctapp_2_per_curiam


class Site(fladistctapp_2_per_curiam.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.type_id = 1  # Written
        self.url = self.get_url()



from juriscraper.OpinionSite import OpinionSite
import time
from datetime import date
from lxml import html

class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'http://www.ca10.uscourts.gov/opinions/new/daily_decisions.rss'
        self.court_id = self.__module__

        # HTTPS certificate is bad, but hopefully they'll fix it and we can remove the line below
        self.disable_certificate_verification()

    def _get_case_names(self):
        return [e for e in self.html.xpath('//item/title/text()')]

    def _get_download_urls(self):
        return [html.tostring(e, method='text') for e in self.html.xpath('//item/link')]

    def _get_case_dates(self):
        dates = []
        for date_string in self.html.xpath('//item/pubdate/text()'):
            date_only = ' '.join(date_string.split(' ')[1:4])
            dates.append(date.fromtimestamp(time.mktime(time.strptime(date_only, '%d %b %Y'))))
        return dates

    def _get_docket_numbers(self):
        return [e.split(':')[1] for e in self.html.xpath('//item/description/text()[1]')]

    def _get_precedential_statuses(self):
        return [e for e in self.html.xpath('//item/category/text()')]

    def _get_lower_courts(self):
        return [e for e in self.html.xpath('//item/description/b/text()')]



# coding=utf-8
"""Scraper for Connecticut Supreme Court
CourtID: conn
Court Short Name: Conn.
Author: Asadullah Baig <asadullahbeg@outlook.com>

History:
 - 2014-07-11: created
 - 2014-08-08, mlr: updated to fix InsanityError on case_dates
 - 2014-09-18, mlr: updated XPath to fix InsanityError on docket_numbers
 - 2015-06-17, mlr: made it more lenient about date formatting
 - 2016-07-21, arderyp: fixed to handle altered site format
 - 2017-01-10, arderyp: restructured to handle new format use case that includes
        opinions without dates and flagged for 'future' publication
"""

from lxml import etree
from datetime import date

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string, normalize_dashes


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.crawl_date = date.today()
        self.url = 'http://www.jud.ct.gov/external/supapp/archiveAROsup{year}.htm'.format(
            year=self.crawl_date.strftime("%y"))
        self.court_id = self.__module__
        self.cases = []

    def _download(self, request_dict={}):
        html = super(Site, self)._download(request_dict)
        self._extract_cases_from_html(html)
        return html

    def _extract_cases_from_html(self, html):
        """Build list of data dictionaries, one dictionary per case (table row)."""
        # Strip inconsistently placed <font> and <br>
        # tags that make stable coverage almost impossible
        etree.strip_tags(html, 'font', 'br')
        path = '//table[@id="AutoNumber1"]//ul'
        for ul in html.xpath(path):
            preceding = ul.xpath('./preceding::*[1]')[0]
            preceding_text = ' '.join(preceding.text_content().split()).strip(':')
            # Skip sections that are marked to be published at future date
            if preceding_text and not preceding_text.lower().endswith(' date'):
                # Below will fail if they change up string format
                date_string = preceding_text.split()[-1]
                case_date = convert_date_string(date_string)
                for element in ul.xpath('./li | ./a'):
                    if element.tag == 'li':
                        text = normalize_dashes(' '.join(element.text_content().split()))
                        if not text:
                            continue
                        anchor = element.xpath('.//a')[0]
                    elif element.tag == 'a':
                        # Malformed html, see connappct_example.html
                        anchor = element
                        glued = '%s %s' % (anchor.text_content(), anchor.tail)
                        text = normalize_dashes(' '.join(glued.split()))
                    self.cases.append({
                        'date': case_date,
                        'url': anchor.xpath('./@href')[0],
                        'docket': text.split('-')[0].replace('Concurrence', '').replace('Dissent', ''),
                        'name': text.split('-', 1)[1],
                    })

    def _get_case_names(self):
        return [case['name'] for case in self.cases]

    def _get_download_urls(self):
        return [case['url'] for case in self.cases]

    def _get_case_dates(self):
        return [case['date'] for case in self.cases]

    def _get_docket_numbers(self):
        return [case['docket'] for case in self.cases]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.cases)



# Auth: mlr
# Date: 2013-04-26

from juriscraper.opinions.united_states.state import miss
from datetime import date
from lxml import html


class Site(miss.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__

        # If it's the beginning of January, we need to make sure that we aren't
        # missing any late-coming cases from the previous year.
        today = date.today()
        self.url = ('http://courts.ms.gov/scripts/websiteX_cgi.exe/GetOpinion?'
                    'Year={year}&Court=Supreme+Court&Submit=Submit').format(
            year=int(today.year) - 1
        )
        beginning_of_year = (
            date(today.year, 1, 1) <= today <= date(today.year, 1, 15))
        if not beginning_of_year:
            # This simply aborts the crawler.
            self.status = 200
            self.html = html.fromstring('<html></html>')



from __future__ import print_function

import getpass
import logging.handlers
import traceback

import sys

LOG_FILENAME = '/var/log/juriscraper/debug.log'

def errprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

def make_default_logger(file_path=LOG_FILENAME):
    """Boilerplate and testing code to create a logger. If we run into an
    IOError, issue a warning and use the NullHandler so things work without
    functional logging.

    :return: a logger object
    """
    logger = logging.getLogger('Logger')
    if not len(logger.handlers):
        logger.setLevel(logging.DEBUG)
        # Create a handler and attach it to the logger
        try:
            handler = logging.handlers.RotatingFileHandler(
                file_path,
                maxBytes=5120000,
                backupCount=7
            )
        except IOError as e:
            if e.errno == 2:
                errprint("\nWarning: %s: %s. " \
                      "Have you created the directory for the log?" % (
                          e.strerror,
                          file_path,
                      ))
            elif e.errno == 13:
                errprint("\nWarning: %s: %s. " \
                      "Cannot access file as user: %s" % (
                          e.strerror,
                          file_path,
                          getpass.getuser(),
                      ))
            else:
                errprint("\nIOError [%s]: %s\n%s" % (
                    e.errno,
                    e.strerror,
                    traceback.format_exc()
                ))
            errprint("Juriscraper will continue to run, and all logs will be "
                  "sent to stderr.")
            handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(
            logging.Formatter('%(asctime)s - %(levelname)s: %(message)s')
        )
        logger.addHandler(handler)
    return logger



"""
Scraper for Florida 5th District Court of Appeal
CourtID: flaapp5
Court Short Name: flaapp5
Court Contact: 5dca@flcourts.org, 386-947-1530
Author: Andrei Chelaru
Reviewer: mlr
History:
 - 2014-07-23, Andrei Chelaru: Created.
 - 2014-08-05, mlr: Updated.
 - 2014-08-06, mlr: Updated.
 - 2014-09-18, mlr: Updated date parsing code to handle Sept.
 - 2016-03-16, arderyp: Updated to return proper absolute pdf url paths, simplify date logic
"""

import re
from lxml import html

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string
from juriscraper.lib.string_utils import normalize_dashes
from juriscraper.lib.string_utils import clean_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.5dca.org/opinions_archived.shtml'
        self.base_path = "//a"
        self.case_regex = '(5D.*-.*\d{1,3})([- ]+[A-Za-z1-9].*)'

    def _download(self, request_dict={}):
        html_l = super(Site, self)._download(request_dict)

        # Test/example files should use html from direct resource page
        # NOTE: fladistctapp_5_example_5.html SHOULD have 0 results
        if self.test_mode_enabled():
            return [html_l]

        html_trees = []
        # this path reads the link of the last 2 dates
        path = "(//a[contains(./@href, 'filings')])[position() < 3]/@href"
        # to get all the dates in that page the following path can be used:
        # path = "//a[contains(./@href, 'filings')]"
        for url in html_l.xpath(path):
            html_tree = self._get_html_tree_by_url(url, request_dict)

            # This district has some nasty HTML that occasionally breaks one
            # case name across two anchors. For example (http://www.5dca.org/Opinions/Opin2014/072114/filings%20072114.html):
            #
            #      <a href='...'>5D12-4340, 5D12-4401 and 5D12-4319</a><br>
            #      <a href='...'>N.R. v. Florida Birth-Related</a>
            #
            # The solution is to look at all the a's, and merge the text when
            # the href's are identical.
            previous_a = None
            for e in html_tree.xpath('//a'):
                if previous_a is not None and e.attrib['href'] == previous_a.attrib['href']:
                    # Same. Merge their text then delete the extraneous element.
                    previous_text = html.tostring(previous_a, method='text', encoding='unicode')
                    e_text = html.tostring(e, method='text', encoding='unicode')
                    previous_a.text = " ".join((previous_text + e_text).split())

                    # See: http://stackoverflow.com/a/7981894/64911
                    e.getparent().remove(e)
                previous_a = e

            html_trees.append(html_tree)
        return html_trees

    def _get_case_names(self):
        case_names = []
        for html_tree in self.html:
            case_names.extend(self._return_case_names(html_tree))
        return case_names

    def _return_case_names(self, html_tree):
        names = []
        for anchor in self._return_anchors_with_text(html_tree):
            text = self.sanitize_text(anchor.text_content().strip())
            names.append(re.search(self.case_regex, text).group(2))
        return names

    def _get_download_urls(self):
        download_urls = []
        for html_tree in self.html:
            download_urls.extend(self._return_download_urls(html_tree))
        return download_urls

    def _return_download_urls(self, html_tree):
        return [a.xpath('@href')[0] for a in self._return_anchors_with_text(html_tree)]

    def _get_case_dates(self):
        case_dates = []
        for html_tree in self.html:
            dates = self._return_dates(html_tree)
            if dates:
                case_dates.extend(dates)
        return case_dates

    def _return_dates(self, html_tree):
        prefixes = ['Opinions', 'OPINIONS', 'Opinion', 'OPINION']
        paths = ["//*[starts-with(., '%s ')]/text()" % string for string in prefixes]
        path = ' | '.join(paths)
        date_text = html_tree.xpath(path)
        # Some pages have only PCAS publications without opinions
        # example: http://www.5dca.org/Opinions/Opin2016/092616/filings%20092616.html
        if not date_text:
            return False
        text = date_text[0].lower()
        date_string = re.search('.* week of (.*)', text).group(1).strip()
        case_date = convert_date_string(date_string)
        return [case_date] * len(self._return_anchors_with_text(html_tree))

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_dates)

    def _get_docket_numbers(self):
        docket_numbers = []
        for html_tree in self.html:
            docket_numbers.extend(self._return_docket_numbers(html_tree))
        return docket_numbers

    def _return_docket_numbers(self, html_tree):
        dockets = []
        for anchor in self._return_anchors_with_text(html_tree):
            text = self.sanitize_text(anchor.text_content().strip())
            docket_raw = re.search(self.case_regex, text).group(1)
            docket_clean = docket_raw.replace('&', '').replace(',', '')
            dockets.append(', '.join(docket_clean.split()))
        return dockets

    # Sometimes the html includes empty anchors,
    # we should only process href for anchors with text
    def _return_anchors_with_text(self, html_tree):
        return [a for a in html_tree.xpath(self.base_path) if a.text_content().strip()]

    def sanitize_text(self, text):
        """Prevent non-standard characters and typos from breaking regex"""
        return self.fix_court_year_id_typo(clean_string(normalize_dashes(text)))

    def fix_court_year_id_typo(self, text):
        """Work around court typo
        String missing leading '5', see fladistctapp_5_example_7.html
        Will return the same exact string 99.99% of the time
        """
        court_year_id = text.split('-')[0]
        if len(court_year_id) == 3 and court_year_id[0] == 'D':
            text = '5%s' % text
        return text



import signal
import six
import sys
import traceback
from six.moves.urllib import parse as six_parse
from six.moves.urllib import request as six_request
from optparse import OptionParser

from juriscraper.lib.importer import build_module_list, site_yielder
from juriscraper.lib.string_utils import trunc

die_now = False


def signal_handler(signal, frame):
    # Trigger this with CTRL+4
    v_print(3, '**************')
    v_print(3, 'Signal caught. Finishing the current court, then exiting...')
    v_print(3, '**************')
    global die_now
    die_now = True


def extract_doc_content(data):
    # Your data extraction routines here. (pdftotext, abiword, etc.)
    return data


def scrape_court(site, binaries=False):
    """Calls the requested court(s), gets its content, then throws it away.

    Note that this is a very basic caller lacking important functionality, such
    as:
     - checking whether the HTML of the page has changed since last visited
     - checking whether the downloaded content is already in your data store
     - saving anything at all

    Nonetheless, this caller is useful for testing, and for demonstrating some
    basic pitfalls that a caller will run into.
    """
    for item in site:
        # Percent encode URLs (this is a Python wart)
        download_url = six_parse.quote(item['download_urls'], safe="%/:=&?~#+!$,;'@()*[]")

        if binaries:
            try:
                opener = six_request.build_opener()
                for cookie_dict in site.cookies:
                    opener.addheaders.append(("Cookie", "%s=%s" % (cookie_dict['name'], cookie_dict['value'])))
                data = opener.open(download_url).read()
                # test for empty files (thank you CA1)
                if len(data) == 0:
                    v_print(3, 'EmptyFileError: %s' % download_url)
                    v_print(3, traceback.format_exc())
                    continue
            except Exception:
                v_print(3, 'DownloadingError: %s' % download_url)
                v_print(3, traceback.format_exc())
                continue

            # Extract the data using e.g. antiword, pdftotext, etc., then
            # clean it up.
            data = extract_doc_content(data)
            data = site.cleanup_content(data)

        # Normally, you'd do your save routines here...
        v_print(1, '\nAdding new item:')
        for k, v in item.items():
            if isinstance(v, six.text_type):
                value = trunc(v, 200, ellipsis='...')
                v_print(1, '    %s: "%s"' % (k, value.encode('utf-8')))
            else:
                # Dates and such...
                v_print(1, '    %s: %s' % (k, v))

    v_print(3, '\n%s: Successfully crawled %d items.' % (site.court_id, len(site)))


v_print = None
def main():
    global die_now

    # this line is used for handling SIGTERM (CTRL+4), so things can die safely
    signal.signal(signal.SIGTERM, signal_handler)

    usage = ('usage: %prog -c COURTID [-d|--daemon] [-b|--binaries]\n\n'
             'To test ca1, downloading binaries, use: \n'
             '    python %prog -c opinions.united_states.federal_appellate.ca1 -b\n\n'
             'To test all federal courts, omitting binaries, use: \n'
             '    python %prog -c opinions.united_states.federal_appellate')
    parser = OptionParser(usage)
    parser.add_option('-c', '--courts', dest='court_id', metavar="COURTID",
                      help=('The court(s) to scrape and extract. This should be in '
                            'the form of a python module or package import '
                            'from the Juriscraper library, e.g. '
                            '"juriscraper.opinions.united_states.federal.ca1" or '
                            'simply "opinions" to do all opinions. If desired, '
                            'you can use slashes instead of dots to separate'
                            'the import path.'))
    parser.add_option('-d', '--daemon', action="store_true", dest='daemonmode',
                      default=False, help=('Use this flag to turn on daemon '
                                           'mode, in which all courts requested '
                                           'will be scraped in turn, non-stop.'))
    parser.add_option('-b', '--download_binaries', action='store_true',
                      dest='binaries',
                      default=False,
                      help=('Use this flag if you wish to download the pdf, '
                            'wpd, and doc files.'))
    parser.add_option('-v',
                      '--verbosity',
                      action='count',
                      default=1,
                      help='Increase output verbosity (e.g., -vv is more than -v).')
    parser.add_option('--backscrape',
                      dest='backscrape',
                      action='store_true',
                      default=False,
                      help='Download the historical corpus using the _download_backwards method.')

    (options, args) = parser.parse_args()

    daemon_mode = options.daemonmode
    binaries = options.binaries
    court_id = options.court_id
    backscrape = options.backscrape

    # Set up the print function
    print("Verbosity is set to: %s" % options.verbosity)
    def _v_print(*verb_args):
        if verb_args[0] > (3 - options.verbosity):
            print(verb_args[1])

    global v_print
    v_print = _v_print

    if not court_id:
        parser.error('You must specify a court as a package or module.')
    else:
        court_id = court_id.replace('/', '.')
        if court_id.endswith('.py'):
            court_id = court_id[:-3]

        module_strings = build_module_list(court_id)
        if len(module_strings) == 0:
            parser.error('Unable to import module or package. Aborting.')

        v_print(3, 'Starting up the scraper.')
        num_courts = len(module_strings)
        i = 0
        while i < num_courts:
            # this catches SIGINT, so the code can be killed safely.
            if die_now:
                v_print(3, 'The scraper has stopped.')
                sys.exit(1)

            package, module = module_strings[i].rsplit('.', 1)
            v_print(3, "Current court: %s.%s" % (package, module))

            mod = __import__("%s.%s" % (package, module),
                             globals(),
                             locals(),
                             [module])
            try:
                if backscrape:
                    for site in site_yielder(mod.Site().back_scrape_iterable, mod):
                        site.parse()
                        scrape_court(site, binaries)
                else:
                    site = mod.Site()
                    v_print(3, 'Sent %s request to: %s' % (site.method, site.url))
                    if site.uses_selenium:
                        v_print(3, "Selenium will be used.")
                    site.parse()
                    scrape_court(site, binaries)
            except Exception:
                v_print(3, '*************!! CRAWLER DOWN !!****************')
                v_print(3, '*****scrape_court method failed on mod: %s*****' % module_strings[i])
                v_print(3, '*************!! ACTION NEEDED !!***************')
                v_print(3, traceback.format_exc())
                i += 1
                continue

            last_court_in_list = (i == (num_courts - 1))
            if last_court_in_list and daemon_mode:
                i = 0
            else:
                i += 1

    v_print(3, 'The scraper has stopped.')
    sys.exit(0)

if __name__ == '__main__':
    main()



# Author: Phil Ardery
# Date created: 2017-01-27
# Contact: 501-682-9400 (Administrative Office of the Curt)

import time
from juriscraper.AbstractSite import logger
from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import titlecase, convert_date_string

## WARNING: THIS SCRAPER IS FAILING:
## This scraper is succeeding in development, but
## is failing in production.  We are not exactly
## sure why, and suspect that the hosting court
## site may be blocking our production IP and/or
## throttling/manipulating requests from production.


class Site(OpinionSite):
    """This website implements anti-bot mechanisms. There is some threshold
    where, after some X number of rapid requests, the service requires a
    catptcha to continue. I have no idea where the threshold is. I think
    it's pretty generous, and shouldn't be triggered during the course of
    normal operations. However, it may be triggered during extensive testing,
    when running the scraper many times in succession.
    """
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.cases = []
        self.subpage_html = False
        self.url_id = 'supremecourt'
        self.url = self.get_url()

    def get_url(self):
        return 'https://opinions.arcourts.gov/ark/%s/en/rss.do' % self.url_id

    def _download(self, request_dict={}):
        """We have to scrape each item url in the RSS feed, since the
        docket number and other info isn't included in the RSS result
        """
        html = super(Site, self)._download(request_dict)
        self.extract_cases_from_subpages(html)
        return html

    def extract_cases_from_subpages(self, html):
        """In order for testing to work, and to prevent hitting the
        network while running tests, each example file should correspond
        to a specific case page, NOT the main RSS feed page. When creating
        a new example file, you can retrieve the html with:
            curl "<CASE-PAGE-URL>?iframe=true"
        """
        if self.test_mode_enabled():
            self.subpage_html = html
            self.extract_case_from_subpage()
        else:
            filter_by_citation = 'contains(./title, "Ark.") or contains(./title, "ARK.")'
            path = '//item[%s]/link' % filter_by_citation
            subpage_urls = [href.tail.strip() for href in html.xpath(path)]
            for i, subpage_url in enumerate(subpage_urls):
                time.sleep(2.5)  # Avoid anti-bot mechanism
                subpage_url = subpage_url + '?iframe=true'
                # cat /var/log/juriscraper/debug.log for more info
                logger.info('%s: Scraping sub-page: %s' % (self.court_id, subpage_url))
                self.subpage_html = self._get_html_tree_by_url(subpage_url)
                self.extract_case_from_subpage()

    def extract_case_from_subpage(self):
        self.cases.append({
            'name': self.subpage_html.xpath('//h3[@class="title"]')[0].text_content().strip(),
            'url': self.subpage_html.xpath('//div[@class="documents"]/a/@href')[0],
            'date': self.get_subpage_value_by_key('Date'),
            'judge': self.get_subpage_value_by_key('Author'),
            'docket': self.get_subpage_value_by_key('Docket Number'),
            'citation': self.get_subpage_value_by_key('Neutral Citation'),
        })

    def get_subpage_value_by_key(self, key):
        """Return value in column 2 where column 1 includes $key.
        If no $key in column 1, return empty string (for example,
        not all records have docket numbers).
        """
        path = '//tr[contains(./td[@class="label"], "%s")]/td[@class="metadata"]' % key
        elements = self.subpage_html.xpath(path)
        return elements[0].text_content().strip() if elements else ''

    def _get_download_urls(self):
        return [case['url'] for case in self.cases]

    def _get_case_names(self):
        return [case['name'] for case in self.cases]

    def _get_case_dates(self):
        return [convert_date_string(case['date']) for case in self.cases]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)

    def _get_docket_numbers(self):
        return [case['docket'] for case in self.cases]

    def _get_judges(self):
        return [case['judge'].lower().title() for case in self.cases]

    def _get_neutral_citations(self):
        return [case['citation'] for case in self.cases]

    def _get_dispositions(self):
        return [titlecase(case['docket']) for case in self.cases]


# Scraper for Florida Supreme Court
# CourtID: fla
# Court Short Name: fla

from datetime import datetime

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        today = datetime.today().strftime('%m/%d/%Y')
        this_year = datetime.today().year
        last_year = this_year - 1
        today_last_year = today.replace(str(this_year), str(last_year))
        # Get 50 most recent opinions from past year--even though you can
        # put whatever number you want as limit, 50 seems to be max
        self.url = 'https://www.floridasupremecourt.org/search/?searchtype=opinions&limit=50&startdate=%s&enddate=%s' % (today_last_year, today)
        self.path_cell = '//div[@class="search-results"]//tr/td[%d]'
        self.cases = []

    def _download(self, request_dict={}):
        html = super(Site, self)._download(request_dict)
        self._extract_cases_from_html(html)
        return html

    def _extract_cases_from_html(self, html):
        path = '//div[@class="search-results"]//tbody/tr'
        for row in html.xpath(path):
            cells = row.xpath('./td')
            url = cells[4].xpath('./a/@href')
            if not url:
                # Skip rows without PDFs
                continue
            self.cases.append({
                'url': url[0],
                'name': cells[2].text_content(),
                'docket': cells[1].text_content(),
                'date': convert_date_string(cells[0].text_content()),
            })

    def _get_case_names(self):
        return [case['name'] for case in self.cases]

    def _get_download_urls(self):
        return [case['url'] for case in self.cases]

    def _get_case_dates(self):
        return [case['date'] for case in self.cases]

    def _get_docket_numbers(self):
        return [case['docket'] for case in self.cases]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.cases)



from . import alaska


class Site(alaska.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'https://appellate-records.courts.alaska.gov/CMSPublic/Home/Opinions?isCOA=True'



"""
Scraper for Indiana Supreme Court
CourtID: ind
Court Short Name: Ind.
Auth: Jon Andersen <janderse@gmail.com>
Reviewer: mlr
History:
    2014-09-03: Created by Jon Andersen
"""
from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import clean_if_py3
import time
from datetime import date


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'http://www.in.gov/judiciary/opinions/supreme.html'
        self.court_id = self.__module__
        self.my_precedential_statuses = []

    def _get_case_names(self):
        raw_case_names = [s for s in self.html.xpath('//dl/dt/a/text()')]
        case_names = []
        self.my_precedential_statuses = []
        for case_name in raw_case_names:
            if case_name.find("(NFP)") >= 0:
                case_names.append(case_name.replace("(NFP)", "").strip())
                self.my_precedential_statuses.append("Unpublished")
            else:
                case_names.append(case_name)
                self.my_precedential_statuses.append("Published")
        return case_names

    def _get_download_urls(self):
        return [s for s in self.html.xpath('//dl/dt/a/@href')]

    def _get_case_dates(self):
        dates = []
        for date_string in self.html.xpath('//dl/dd/dd/dd/text()'):
            date_string = clean_if_py3(date_string).strip()
            if date_string == '':
                dates.append('')
            else:
                dates.append(date.fromtimestamp(
                    time.mktime(time.strptime(date_string, '%m/%d/%y'))))
        return dates

    def _get_docket_numbers(self):
        return [s for s in self.html.xpath('//dl/dd/text()')]

    def _get_lower_court_numbers(self):
        return [e if e.strip() != "N/A" else "" for e in self.html.xpath('//dl/dd/dd/text()')]

    def _get_precedential_statuses(self):
        return self.my_precedential_statuses



#!/usr/bin/env python
#
#  Takes an .html file on the command line, parses it using the PACER
#  Docket Report parser, and outputs json to stdout.

import jsondate as json
import sys

from juriscraper.pacer.http import PacerSession
from juriscraper.pacer import DocketReport

pacer_session = PacerSession(username='tr1234',
                             password='Pass!234')
report = DocketReport('psc', pacer_session)

for path in sys.argv[1:]:
    with open(path, 'r') as f:
        report._parse_text(f.read().decode('utf-8'))
    data = report.data
    print json.dumps(data, indent=2, sort_keys=True, separators=(',', ': '))



"""Scraper for Ninth Circuit of Appeals
CourtID: ca9
Court Short Name: ca9
Author: Andrei Chelaru
Reviewer: mlr
Date created: 20 July 2014
History:
 - 2014-11-04: Updated by mlr to account for (1) varied audio file locations,
               (2) Finding the latest 100 items, (3) Handling bad audio URLs.
"""

from datetime import datetime

from juriscraper.DeferringList import DeferringList
from juriscraper.OralArgumentSite import OralArgumentSite
from juriscraper.lib.exceptions import InsanityException
from juriscraper.lib.string_utils import titlecase


class Site(OralArgumentSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'https://www.ca9.uscourts.gov/media/'
        self._set_parameters()
        self.method = 'POST'
        # One less than total number of pages, because last page has a bit of a
        # mess, and it's easier to simply not deal with it.
        self.back_scrape_iterable = range(1, 268)

    def _set_parameters(self, page=1):
        self.parameters = {
            'c_mode': 'view',
            'c_page_size': '100',
            'c_sort_field': '12',
            'c_sort_field_by': '7',
            'c_sort_type': 'desc',
            'c_field_type': '',
            'c_p': str(page),
        }

    def _post_parse(self):
        """Unfortunately, some of the items do not have audio files despite
        appearing in the table and having a link to a supplementary audio page.

        For these items, we set the download_url to '' and this method finds
        the related information for those items and then removes it from all
        the other attributes for the Site object.
        """
        # Start by checking sanity. This will make sure we don't mess things
        # up. If this sanity check fails, we'll know things were messed up
        # before we began tinkering with them.
        self._check_sanity()

        # Items are purged in two steps. First, we identify the index of the
        # items that need purging.
        purge_indexes = []
        for i, url in enumerate(self.download_urls):
            if not url:
                purge_indexes.append(i)

        # Quick check: We did find *some* urls, right?
        if len(purge_indexes) == len(self.download_urls):
            raise InsanityException("Didn't get any download URLs. Looks like "
                                    "something is wrong in the _post_parse() "
                                    "method.")

        # Second, we purge them, beginning at the end and moving forwards. This
        # ensures that we don't delete the wrong items.
        for index_to_purge in sorted(purge_indexes, reverse=True):
            for attr in self._all_attrs:
                item = getattr(self, attr)
                if item is not None:
                    # If we've added stuff to it, then delete the key.
                    del item[index_to_purge]

    def _get_download_urls(self):
        """Links from the root page go to a second page where the real links
        are posted.
        """
        def fetcher(seed_url):
            if self.method == 'LOCAL':
                return "No links fetched during tests."
            else:
                # Goes to second page, grabs the link and returns it.
                html_tree = self._get_html_tree_by_url(seed_url)
                path_to_audio_file = "//*[@class='padboxauto_MediaContent']//a/@href"
                try:
                    url = html_tree.xpath(path_to_audio_file)[0]
                except IndexError:
                    # The URL wasn't found, so something is wrong and we'll have to
                    # fix it in the _post_parse() method.
                    url = ''
                return url

        path = "//tr[@class='dg_tr']/td[6]//@href"
        seed_urls = self.html.xpath(path)
        return DeferringList(seed=seed_urls, fetcher=fetcher)

    def _get_download_urls_orig(self):
        """Note that the links from the root page go to a second page, where
        the real links are posted.
        """
        row_path = "//tr[@class='dg_tr'][.//*[contains(@id, 'case_num')]]"
        date_path = ".//*[contains(concat(' ',@id,' '),' mod_hearing_date')]/text()"
        docket_path = ".//*[contains(concat(' ',@id,' '),' case_num')]/text()"
        urls = []
        for row in self.html.xpath(row_path):
            date_string = row.xpath(date_path)[0]
            d = datetime.strptime(date_string, '%m/%d/%Y').date()
            docket_string = row.xpath(docket_path)[0]
            urls.append(self._return_download_url(d, docket_string))
        return urls

    @staticmethod
    def _return_download_url(d, docket_string):
        link = "http://cdn.ca9.uscourts.gov/datastore/media/{date}/{docket_nr}.wma".format(
            date=d.strftime('%Y/%m/%d'),
            docket_nr=docket_string,
        )
        return link

    def _get_case_names(self):
        path = "//*[contains(concat(' ',@id,' '),' case_name')]/text()"
        return list(self.html.xpath(path))

    def _get_case_dates(self):
        path = "//*[contains(concat(' ',@id,' '),' mod_hearing_date')]/text()"
        return [datetime.strptime(date_string, '%m/%d/%Y').date()
                for date_string in self.html.xpath(path)]

    def _get_judges(self):
        path = "//*[contains(concat(' ',@id,' '),' case_panel')]/text()"
        return [titlecase(s.lower()) for s in self.html.xpath(path)]

    def _get_docket_numbers(self):
        path = "//*[contains(concat(' ',@id,' '),' case_num')]/text()"
        return list(self.html.xpath(path))

    def _download_backwards(self, page):
        self._set_parameters(page=page)
        self.html = self._download()



from juriscraper.OpinionSite import OpinionSite
import re
import time
from datetime import date
from dateutil.rrule import MONTHLY, rrule


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        today = date.today()
        self.url = 'http://media.ca8.uscourts.gov/cgi-bin/opnByMM.pl?theMM=%02d&theYY=%s&A1=Get+Opinions' % (today.month, today.year)
        self.court_id = self.__module__

        self.back_scrape_iterable = [i.date() for i in rrule(
            MONTHLY,
            dtstart=date(1995, 11, 1),
            until=date(2015, 1, 1),
        )]

    def _get_case_names(self):
        case_names = []
        case_name_regex = re.compile('(\d{2}/\d{2}/\d{4})(.*)')
        for text in self.html.xpath('//a[contains(@href, "opndir")]/following-sibling::b/text()'):
            case_names.append(case_name_regex.search(text).group(2))
        return case_names

    def _get_download_urls(self):
        return [e for e in self.html.xpath('//a[contains(@href, "opndir")]/@href')]

    def _get_case_dates(self):
        case_dates = []
        case_date_regex = re.compile('(\d{2}/\d{2}/\d{4})(.*)')
        for text in self.html.xpath('//a[contains(@href, "opndir")]/following-sibling::b/text()'):
            date_string = case_date_regex.search(text).group(1)
            case_dates.append(date.fromtimestamp(time.mktime(time.strptime(date_string, '%m/%d/%Y'))))
        return case_dates

    def _get_docket_numbers(self):
        docket_numbers = []
        docket_number_regex = re.compile('(\d{2})(\d{4})(u|p)', re.IGNORECASE)
        for docket_number in self.html.xpath('//a[contains(@href, "opndir")]/text()'):
            regex_results = docket_number_regex.search(docket_number)
            docket_numbers.append('%s-%s' % (regex_results.group(1), regex_results.group(2)))
        return docket_numbers

    def _get_precedential_statuses(self):
        statuses = []
        for docket_number in self.html.xpath('//a[contains(@href, "opndir")]/text()'):
            docket_number = docket_number.split('.')[0]
            if 'p' in docket_number.lower():
                statuses.append('Published')
            elif 'u' in docket_number.lower():
                statuses.append('Unpublished')
            else:
                statuses.append('Unknown')
        return statuses

    def _download_backwards(self, d):

        self.url = 'http://media.ca8.uscourts.gov/cgi-bin/opnByMM.pl?theMM=%02d&theYY=%s&A1=Get+Opinions' % (
            d.month,
            d.year
        )

        self.html = self._download()
        if self.html is not None:
            # Setting status is important because it prevents the download
            # function from being run a second time by the parse method.
            self.status = 200



from juriscraper.opinions.united_states.state import washctapp_p


class Site(washctapp_p.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.courtLevel = 'C'
        self.pubStatus = 'UNP'
        self._set_parameters()

    def _get_precedential_statuses(self):
        return ["Unpublished"] * len(self.case_names)




# coding=utf-8

from juriscraper.OpinionSiteLinear import OpinionSiteLinear
from juriscraper.lib.exceptions import InsanityException
from juriscraper.lib.string_utils import clean_string


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.isc.idaho.gov/appeals-court/coaunpublished'
        self.status = 'Unpublished'

    def _process_html(self):
        for item in self.html.xpath('//li[contains(.//a/@href, ".pdf")]'):
            text = clean_string(item.text_content())
            text_parts = text.split('-', 1)

            if len(text_parts) != 2:
                raise InsanityException('Unexpected text format: "%s"' % text)

            # sometimes the records include a docket number(s) as the
            # first words in the second half of the hyphenated string,
            # but some don't include a docket at all.  So we test to see
            # if the first word is numeric (minus the slash characters
            # used to conjoin multiple docket numbers).
            docket_name = text_parts[1].split(None, 1)
            first_word = docket_name[0].replace('/', '')
            if first_word.isnumeric():
                docket = docket_name[0]
                name = docket_name[1]
            else:
                docket = ''
                name = text_parts[1]

            self.cases.append({
                'date': text_parts[0],
                'docket': docket,
                'name': name,
                'url': item.xpath('.//a/@href')[0],
            })



"""Scraper for the Supreme Court of Ohio
CourtID: ohio
Court Short Name: Ohio
Author: Andrei Chelaru
Reviewer: mlr
History:
 - Stubbed out by Brian Carver
 - 2014-07-30: Finished by Andrei Chelaru
 - 2015-07-31: Redone by mlr to use ghost driver. Alas, their site used to be
               great, but now it's terribly frustrating.
"""
import os
from datetime import date, datetime

from lxml import html
from lxml.html import tostring
from selenium import webdriver

from juriscraper.AbstractSite import logger, phantomjs_executable_path
from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.html_utils import fix_links_in_lxml_tree
from juriscraper.lib.string_utils import clean_if_py3


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        # Changing the page # in the url will get additional pages
        # Changing the source # (0-13) will get the 12 Courts of Appeals and
        # the Court of Claims. We do not use the "all sources" link because a
        # single day might yield more than 25 opinions and this scraper is
        # not designed to walk through multiple pages.
        self.court_index = 0
        self.year = str(date.today().year)
        self.url = 'http://www.supremecourtofohio.gov/rod/docs/'
        self.court_id = self.__module__
        self.base_path = "id('MainContent_gvResults')//tr[position() > 1]/td[2][string-length(normalize-space(text())) > 1]"
        self.uses_selenium = True

    def _download(self, request_dict={}):
        """This is another of the cursed MS asp.net pages with damned POST
          parameters like __EVENTVALIDATION. These are near impossible to
          scrape without using Selenium.
        """
        if self.test_mode_enabled():
            return super(Site, self)._download(request_dict=request_dict)

        driver = webdriver.PhantomJS(
            executable_path=phantomjs_executable_path,
            service_log_path=os.path.devnull,  # Disable ghostdriver.log
        )
        driver.implicitly_wait(30)
        logger.info("Now downloading case page at: %s" % self.url)
        driver.get(self.url)

        # Court drop down...
        driver.find_element_by_xpath(
            "//select[@id='MainContent_ddlCourt']"
            "/option[@value='{court}']".format(court=self.court_index)
        ).click()

        # Year drop down...
        yearDropDownId = 'MainContent_ddlDecidedYear'
        yearDropDownPath = "//select[@id='{id}']/option[@value='%s']" % self.year
        yearDropdownPaths = [
            yearDropDownPath.format(id=yearDropDownId),             # Legacy examples
            yearDropDownPath.format(id=yearDropDownId + 'Min'),     # current (2017)
        ]
        driver.find_element_by_xpath(' | '.join(yearDropdownPaths)).click()

        # Hit submit
        submitPath = "//input[@id='MainContent_btnSubmit']"
        driver.find_element_by_xpath(submitPath).click()

        # Selenium doesn't give us the actual code, we have to hope.
        self.status = 200

        text = self._clean_text(driver.page_source)
        html_tree = html.fromstring(text)
        html_tree.rewrite_links(fix_links_in_lxml_tree,
                                base_href=self.request['url'])
        return html_tree

    def _get_case_names(self):
        path = "{base}/preceding::td[1]".format(base=self.base_path)
        case_names = []
        for e in self.html.xpath(path):
            case_names.append(
                tostring(e, method='text', encoding='unicode').strip()
            )
        return case_names

    def _get_download_urls(self):
        path = "{base}/preceding::td[1]//a[1]/@href".format(base=self.base_path)
        return list(self.html.xpath(path))

    def _get_docket_numbers(self):
        path = "{base}//text()".format(base=self.base_path)
        return list(self.html.xpath(path))

    def _get_summaries(self):
        path = "{base}/following::td[1]//text()".format(base=self.base_path)
        return list(self.html.xpath(path))

    def _get_case_dates(self):
        path = "{base}/following::td[4]//text()".format(base=self.base_path)
        dates = []
        for s in self.html.xpath(path):
            dates.append(datetime.strptime(clean_if_py3(s).strip(), '%m/%d/%Y').date())
        return dates

    def _get_neutral_citations(self):
        path = "{base}/following::td[6]//text()".format(base=self.base_path)
        return [s.replace('-', ' ') for s in self.html.xpath(path)]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_names)

    def _get_judges(self):
        path = "{base}/following::td[2]".format(base=self.base_path)
        return map(self._return_judge, self.html.xpath(path))

    @staticmethod
    def _return_judge(e):
        txt = e.xpath(".//text()")
        if txt:
            return txt[0]
        else:
            return ''



"""Scraper for the United States Court of Appeals for the Armed Forces
CourtID: armfor
Court Short Name: C.A.A.F."""

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.html_utils import get_html5_parsed_text
from juriscraper.lib.string_utils import convert_date_string


# This court has some funky html that varies
# across the old term listings. Before committing
# changes, make sure you haven't broken the backscraper
# by running: python juriscraper/sample_caller.py -c opinions.united_states.federal_special.armfor --backscrape
class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.back_scrape_iterable = ['']
        self.url = "http://www.armfor.uscourts.gov/newcaaf/opinions.htm"
        self.row_base_path = '//table[@border="1"]//tr[descendant::a]'
        self.path_to_landing_page_links = '//blockquote/ul/li[1]/font/a[1]/@href'

    def _download(self, request_dict={}):
        landing_page_html = super(Site, self)._download(request_dict)

        # Example test files should include html of direct resource page
        if self.test_mode_enabled():
            return [landing_page_html]

        urls = landing_page_html.xpath(self.path_to_landing_page_links)
        return [self._get_html_tree_by_url(url, request_dict) for url in urls]

    def _get_case_names(self):
        names = []
        path = '%s/td[1]' % self.row_base_path
        for html_tree in self.html:
            names.extend([cell.text_content().strip() for cell in html_tree.xpath(path)])
        return names

    def _get_download_urls(self):
        urls = []
        path = '%s/td[2]//a[last()]/@href' % self.row_base_path
        for html_tree in self.html:
            urls.extend([url for url in html_tree.xpath(path)])
        return urls

    def _get_case_dates(self):
        dates = []
        path = '%s/td[3]' % self.row_base_path
        for html_tree in self.html:
            for cell in html_tree.xpath(path):
                dates.append(convert_date_string(cell.text_content()))
        return dates

    def _get_docket_numbers(self):
        dockets = []
        path = '%s//td[2]' % self.row_base_path
        for html_tree in self.html:
            for cell in html_tree.xpath(path):
                docket_raw = cell.text_content().strip()
                dockets.append(docket_raw.rstrip('(PDF)'))
        return dockets

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_names)

    def _download_backwards(self, _):
        # We skip the 1998 and 1999 entries because they
        # are formatted differently and don't contain dates
        limitation = 'not(contains(@href, "1998Term.htm")) and not(contains(@href, "1997Term.htm"))'
        self.path_to_landing_page_links = '//blockquote/ul/li/font/a[1][%s]/@href' % limitation
        self.html = self._download()

    def _make_html_tree(self, text):
        return get_html5_parsed_text(text)



# -*- coding: utf-8 -*-

"""Scraper for the United States Court of Federal Claims
CourtID: uscfc
Court Short Name: Fed. Cl.

Notes:
    Scraper adapted for new website as of February 20, 2014.
"""

import re
import six
import datetime
from lxml import html
from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.exceptions import InsanityException
from juriscraper.lib.string_utils import titlecase, clean_if_py3, convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'http://www.uscfc.uscourts.gov/aggregator/sources/8'
        self.back_scrape_iterable = range(1, 4)
        self.court_id = self.__module__
        self.today = datetime.datetime.now()

    def _download(self, request_dict={}):
        if self.test_mode_enabled():
            # Use static 'today' date for consisting test results
            self.today = convert_date_string('2018/10/17')
        return super(Site, self)._download(request_dict)

    def _get_case_dates(self):
        dates = []
        for item in self.html.xpath('//span[@class="feed-item-date"]'):
            text = item.text_content().strip()
            words = text.split()
            if len(words) == 2:
                date = convert_date_string(words[1])
            elif 'ago' in text:
                # The record was added today "X hours and Y min ago"
                date = self.today
            else:
                raise InsanityException('Unrecognized date element string: %s' % text)
            dates.append(date)
        return dates

    def _get_case_names(self):
        case_names = []
        for t in self.html.xpath('//h3[@class="feed-item-title"]//text()'):
            t = ' '.join(clean_if_py3(t).split())  # Normalize whitespace
            if t.strip():
                # If there is something other than whitespace...
                if not isinstance(t, six.string_types):
                    t = str(t, encoding='utf-8')

                if u' • ' in t:
                    t = t.split(u' • ')[1].strip()
                t = titlecase(t.lower())
                case_names.append(t)
        return case_names

    def _get_download_urls(self):
        path = '//h3[@class="feed-item-title"]/a/@href'
        return list(self.html.xpath(path))

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_names)

    def _get_docket_numbers(self):
        docket_numbers = []
        for t in self.html.xpath('//h3[@class="feed-item-title"]//text()'):
            t = clean_if_py3(t)
            if t.strip():
                # If there is something other than whitespace...
                if not isinstance(t, six.string_types):
                    t = str(t, encoding='utf-8')

                if u' • ' in t:
                    t = t.split(u' • ')[0].strip()
                docket_numbers.append(t)
        return docket_numbers

    def _get_summaries(self):
        summaries = []
        path = '//div[@class="feed-item-body"]'
        for e in self.html.xpath(path):
            s = html.tostring(e, method='text', encoding='unicode')
            s = clean_if_py3(s).split('Keywords:')[0]
            summaries.append(s)

        return summaries

    def _get_judges(self):
        path = '//div[@class="feed-item-body"]'
        judges = []
        splitters = [
            'Signed by Chief Judge',
            'Signed by Judge',
            'Signed by Chief Special Master',  # Vaccine courts have odd names for judges
            'Signed by Special Master',
        ]
        for e in self.html.xpath(path):
            t = html.tostring(e, method='text', encoding='unicode')
            t = clean_if_py3(t).split('Keywords:')[0]
            for splitter in splitters:
                judge_parts = t.rsplit(splitter)
                if len(judge_parts) == 1:
                    # No splits found...
                    judge = ''
                    continue
                else:
                    judge = judge_parts[1]
                    break

            # Often the text looks like: 'Judge Susan G. Braden. (jt1) Copy to parties.' In that case we only
            # want the name, not the rest.
            length_of_match = 2
            m = re.search(r'[a-z]{%s}\.' % length_of_match, judge)  # Two lower case letters followed by a period
            if m:
                judge = judge[:m.start() + length_of_match]
            else:
                judge = ''
            judge.strip('.')
            judges.append(judge)
        return judges

    def _download_backwards(self, page):
        self.url = 'http://www.uscfc.uscourts.gov/aggregator/sources/8?page=%s' % page
        self.html = self._download()




# coding=utf-8
"""Classes for querying PACER's Written Opinion Report (WrtOpRpt.pl),
which is free.
"""

from .reports import BaseReport
from ..lib.date_utils import make_date_range_tuples
from ..lib.html_utils import clean_html, fix_links_in_lxml_tree, \
    get_html_parsed_text, set_response_encoding
from ..lib.log_tools import make_default_logger
from ..lib.string_utils import convert_date_string
from ..pacer.utils import get_nonce_from_form, \
    get_pacer_case_id_from_nonce_url,  get_pacer_doc_id_from_doc1_url, \
    reverse_goDLS_function

logger = make_default_logger()


class FreeOpinionReport(BaseReport):
    """An object for querying and parsing the free opinion report."""

    EXCLUDED_COURT_IDS = ['casb', 'innb', 'miwb', 'ohsb']
    VALID_SORT_PARAMS = ('date_filed', 'case_number')

    def __init__(self, court_id, pacer_session):
        self.responses = []
        self.trees = []
        super(FreeOpinionReport, self).__init__(court_id, pacer_session)

    @property
    def url(self):
        if self.court_id == 'ohnd':
            return 'https://ecf.ohnd.uscourts.gov/cgi-bin/OHND_WrtOpRpt.pl'
        else:
            return ('https://ecf.%s.uscourts.gov/cgi-bin/WrtOpRpt.pl' %
                    self.court_id)

    def query(self, start, end, sort='date_filed', day_span=7):
        """Query the Free Opinions report one day at a time.

        :param start: a date object representing the date you want to start at.
        :param end: a date object representing the date you want to end at.
        :param sort: the order you wish the results to be in, either
        `date_filed` or `case_number`.
        :param day_span: The number of days to query at a time. Defaults to one
        week.
        """
        if self.court_id in self.EXCLUDED_COURT_IDS:
            logger.error("Cannot get written opinions report from '%s'. It is "
                         "not provided by the court or is in disuse.",
                         self.court_id)
            return

        dates = make_date_range_tuples(start, end, gap=day_span)
        responses = []
        for _start, _end in dates:
            _start = _start.strftime('%m/%d/%Y')
            _end = _end.strftime('%m/%d/%Y')
            # Iterate one day at a time. Any more and PACER chokes.
            logger.info("Querying written opinions report for '%s' between %s "
                        "and %s, ordered by %s",
                        self.court_id, _start, _end, sort)

            # Get the first page, grab the nonce, and submit using that.
            response = self.session.get(self.url)
            nonce = get_nonce_from_form(response)
            logger.info("Got nonce of %s", nonce)

            data = {
                'filed_from': _start,
                'filed_to': _end,
                'ShowFull': '1',
                'Key1': self._normalize_sort_param(sort),
                'all_case_ids': '0'
            }
            response = self.session.post(self.url + '?' + nonce, data=data)
            responses.append(response)

        self.responses = responses
        self.parse()

    def parse(self):
        """Using a list of responses, parse out useful information and return
        it as a list of dicts.
        """
        # Reset self.trees before each run or successive runs will add more and
        # more rows.
        self.trees = []
        for response in self.responses:
            response.raise_for_status()
            set_response_encoding(response)
            text = clean_html(response.text)
            tree = get_html_parsed_text(text)
            tree.rewrite_links(fix_links_in_lxml_tree, base_href=response.url)
            self.trees.append(tree)

    @property
    def data(self):
        results = []
        for tree in self.trees:
            opinion_count = int(
                tree.xpath('//b[contains(text(), "Total number of '
                           'opinions reported")]')[0].tail)
            if opinion_count == 0:
                continue
            rows = tree.xpath('(//table)[1]//tr[position() > 1]')
            for row in rows:
                if results:
                    # If we have results already, pass the previous result to
                    # the FreeOpinionRow object.
                    row = FreeOpinionRow(row, results[-1], self.court_id)
                else:
                    row = FreeOpinionRow(row, {}, self.court_id)
                results.append(row)
        logger.info("Parsed %s results from written opinions report at %s",
                    len(results), self.court_id)
        return results

    def _normalize_sort_param(self, sort):
        if sort == 'date_filed':
            return 'de_date_filed'
        elif sort == 'case_number':
            return 'cs_sort_case_numb'
        else:
            raise ValueError("Invalid sort parameter. Value must be one of: %s"
                             % ', '.join(self.VALID_SORT_PARAMS))


class FreeOpinionRow(object):
    """A row in the Free Opinions report.

    For the most part this is fairly straightforward, however eight courts have
    a different type of report that only has four columns instead of the usual
    five (hib, deb, njb, ndb, ohnb, txsb, txwb, vaeb), and a couple courts
    (areb & arwb) have five columns, but are designed more like the four column
    variants.

    The second complication is that cells in the row change position based on
    the sort order of the report. If the report is ordered by date, then the
    first cell is the date. If the report is ordered by case number, then the
    first cell is the case number and case name.

    In general, what we do is detect the column count, and sort order early on
    and then work from there.
    """
    def __init__(self, element, last_good_row, court_id):
        """Initialize the object.

        last_good_row should be a dict representing the values from
        the previous row in the table. This is necessary because the
        report skips the case name if it's the same for two cases in a
        row. For example:

        Joe v. Volcano | 12/31/2008 | 128 | The first doc from case | More here
                       | 12/31/2008 | 129 | The 2nd doc from case   | More here

        By having the values from the previous row, we can be sure to be able
        to complete the empty cells.

        """
        super(FreeOpinionRow, self).__init__()
        self.element = element
        self.last_good_row = last_good_row
        self.court_id = court_id
        self._column_count = self._get_column_count()
        self._sort_order = self._detect_sort_order()

        # Parsed data
        results = self.get_pacer_case_id_and_pacer_seq_no()
        self.pacer_case_id, self.pacer_seq_no = results[0], results[1]
        self.docket_number = self.get_docket_number()
        self.case_name = self.get_case_name()
        self.date_filed = self.get_date_filed()
        self.pacer_doc_id = self.get_pacer_doc_id()
        self.document_number = self.get_document_number()
        self.description = self.get_description()
        self.nature_of_suit = self.get_nos()
        self.cause = self.get_cause()

    def __str__(self):
        return '<FreeOpinionRow in %s>\n%s' % (self.court_id, {
            'pacer_case_id': self.pacer_case_id,
            'docket_number': self.docket_number,
            'case_name': self.case_name,
            'date_filed': self.date_filed,
            'pacer_doc_id': self.pacer_doc_id,
            'document_number': self.document_number,
        })

    def _get_column_count(self):
        return len(self.element.xpath('./td'))

    def _detect_sort_order(self):
        """Detect whether the report is ordered by case number or by date filed.

        If the report is ordered by date filed, you'll have a table row like:

            12/31/2008 | Joe v. Volcano | 128 | The first doc | More here

        If it's ordered by case name, it'll be like:

            Joe v. Volcano | 12/31/2008 | 128 | The first doc | More here

        The case name is always a link, so the simple way to do this
        is to check if there's a link in the second cell of the
        row. If so, it's ordered by date_filed. Else, by
        case_number. Note that the first cell is often blank.
        """
        if len(self.element.xpath('./td[2]//@href')) > 0:
            return 'date_filed'
        else:
            return 'case_number'

    def get_pacer_case_id_and_pacer_seq_no(self):
        """Get the pacer_case_id and pacer_seq_no values for a document.

        Both come from the link to the document, if it has an onclick
        attribute.
        """
        # It's tempting to get this value from the URL in the first
        # cell, but that URL can sometimes differ from the URL used in
        # the goDLS function.  When that's the case, the download
        # fails.
        try:
            onclick = self.element.xpath('./td[3]//@onclick')[0]
        except IndexError:
            pass
        else:
            if 'goDLS' in onclick:
                # Sometimes the onclick is something else, like in insb's free
                # opinion report.
                go_dls_parts = reverse_goDLS_function(onclick)
                return go_dls_parts['caseid'], go_dls_parts['de_seq_num']

        # No onclick, onclick isn't a goDLS link, etc. Try second format.
        if self._sort_order == 'case_number':
            try:
                # This tends to work in the bankr. courts.
                href = self.element.xpath('./td[1]//@href')[0]
            except IndexError:
                logger.info("No content provided in first cell of row. Using "
                            "last good row for pacer_case_id, docket_number, "
                            "and case_name.")
                last_row = self.last_good_row
                return last_row.pacer_case_id, last_row.pacer_seq_no
        elif self._sort_order == 'date_filed':
            href = self.element.xpath('./td[2]//@href')[0]
        return get_pacer_case_id_from_nonce_url(href), None

    def get_docket_number(self):
        try:
            if self._sort_order == 'case_number':
                cell = self.element.xpath('./td[1]//a')[0]
            else:
                cell = self.element.xpath('./td[2]//a')[0]
        except IndexError:
            # No content in the cell.
            return self.last_good_row.docket_number
        else:
            s = cell.text_content().strip()

        if self._column_count == 4 or self.court_id in ['areb', 'arwb']:

            # In this case s will be something like:
            #   14-90018 Stewart v. Kauanui
            # Split on the first space, left is docket number, right
            # is case name.
            return s.split(' ', 1)[0]
        else:
            return s

    def get_case_name(self):
        if self._sort_order == 'case_number':
            cell = self.element.xpath('./td[1]')[0]
        else:
            cell = self.element.xpath('./td[2]')[0]
        s = cell.text_content().strip()
        if not s:
            return self.last_good_row.case_name

        if self._column_count == 4 or self.court_id in ['areb', 'arwb']:
            # See note in docket number
            try:
                return s.split(' ', 1)[1]
            except IndexError:
                # No case name, but a docket number is provided.
                return "Case name unknown"
        else:
            try:
                return cell.xpath('.//b')[0].text_content()
            except IndexError:
                logger.warn("Unable to get case name for %s in %s.",
                            self.docket_number,
                            self.court_id)
                return "Case name unknown"

    def get_date_filed(self):
        if self._sort_order == 'case_number':
            path = './td[2]//text()'
        elif self._sort_order == 'date_filed':
            path = './td[1]//text()'
        s = self.element.xpath(path)[0]
        if not s.strip() and self._sort_order == 'date_filed':
            # Empty cell, return the previous value.
            return self.last_good_row.date_filed
        else:
            return convert_date_string(s)

    def get_pacer_doc_id(self):
        doc1_url = self.element.xpath('./td[3]//@href')[0]
        return get_pacer_doc_id_from_doc1_url(doc1_url)

    def get_document_number(self):
        return self.element.xpath('./td[3]//text()')[0]

    def get_description(self):
        return self.element.xpath('./td[4]')[0].text_content()

    def get_nos(self):
        if self._column_count == 4:
            return ''
        try:
            return self.element.xpath('./td[5]/i[contains(./text(), '
                                      '"NOS")]')[0].tail.strip()
        except IndexError:
            return ''

    def get_cause(self):
        if self._column_count == 4:
            return ''
        try:
            return self.element.xpath('./td[5]/i[contains(./text(), '
                                      '"Cause")]')[0].tail.strip()
        except IndexError:
            return ''



from juriscraper.OpinionSite import OpinionSite
from datetime import datetime
import re


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.courts.state.va.us/scndex.htm'

    def _get_case_names(self):
        path = "//p[./a[contains(./@href, '.pdf')]]/b/text()"
        return list(self.html.xpath(path))

    def _get_docket_numbers(self):
        path = "//p[./a[contains(./@href, '.pdf')]]/a/text()"
        return [doc.strip() for doc in self.html.xpath(path) if doc.strip().isdigit()]

    def _get_case_dates(self):
        dates = []
        path = "//p[./a[contains(./@href, '.pdf')]]/text()"
        pattern = r"^\s*\d{2}/\d{2}/\d{4}"
        for s in self.html.xpath(path):
            date_str = re.findall(pattern, s)
            if len(date_str):
                dates.append(datetime.strptime(date_str[0].strip(), '%m/%d/%Y').date())
        return dates

    def _get_download_urls(self):
        path = "//p[./a[contains(./@href, '.pdf')]]/a[2]/@href"
        urls = [url for url in self.html.xpath(path) if url.endswith('.pdf')]
        return urls

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)



# -*- coding: utf-8 -*-

from juriscraper.lib.string_utils import convert_date_string
from juriscraper.OpinionSite import OpinionSite


class OpinionSiteLinear(OpinionSite):
    """This class can be used for any site that needs to be scraped linearly,
    instead of, for example, with separate html path parsing getters. Sometimes
    it is just easier and less repetitive to scrape a site this way, in which
    case you can simply extend this class and implement _process_html().
    """

    def __init__(self, *args, **kwargs):
        super(OpinionSiteLinear, self).__init__(*args, **kwargs)
        self.cases = []
        self.status = None

    def _process_html(self):
        raise Exception('Must implement _process_html() on OpinionSiteLinear child')

    def _get_case_names(self):
        return [case['name'] for case in self.cases]

    def _get_download_urls(self):
        return [case['url'] for case in self.cases]

    def _get_case_dates(self):
        return [convert_date_string(case['date']) for case in self.cases]

    def _get_precedential_statuses(self):
        if not self.status:
            raise Exception("Must define self.status in __init__ on OpinionSiteLinear child")
        return [self.status] * len(self.cases)

    def _get_docket_numbers(self):
        return [case['docket'] for case in self.cases]

    # optional getters below

    def _get_optional_field_by_id(self, id):
        if self.cases and id in self.cases[0]:
            return [case[id] for case in self.cases]

    def _get_judges(self):
        return self._get_optional_field_by_id('judge')

    def _get_neutral_citations(self):
        return self._get_optional_field_by_id('neutral_citation')



from juriscraper.opinions.united_states.state import nj


class Site(nj.Site):
    """Human web interface: http://www.judiciary.state.nj.us/attorneys/opinions.html#Unpublished_Appellate"""
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.case_types = ['Published_Appellate', 'Unpublished_Appellate']


import os
import hashlib
import json
from datetime import date, datetime
import warnings

import certifi
import requests
import six
from requests.adapters import HTTPAdapter

from juriscraper.lib.date_utils import fix_future_year_typo, json_date_handler
from juriscraper.lib.exceptions import InsanityException
from juriscraper.lib.html_utils import (clean_html, fix_links_in_lxml_tree,
                                        get_html_parsed_text,
                                        set_response_encoding)
from juriscraper.lib.log_tools import make_default_logger
from juriscraper.lib.string_utils import (CaseNameTweaker, clean_string,
                                          harmonize, trunc)
from juriscraper.lib.test_utils import MockRequest

logger = make_default_logger()
phantomjs_executable_path = '/usr/local/bin/phantomjs'
_legacy_path = "/usr/local/phantomjs/phantomjs"
if os.path.isfile(_legacy_path) and os.access(_legacy_path, os.X_OK):
    phantomjs_executable_path = _legacy_path
    msg = """Please place phantomjs executable in /usr/local/bin.
    See https://github.com/freelawproject/juriscraper/pull/241"""
    warnings.warn(msg, DeprecationWarning)


class AbstractSite(object):
    """Contains generic methods for scraping data. Should be extended by all
    scrapers.

    Should not contain lists that can't be sorted by the _date_sort function."""

    def __init__(self, cnt=None):
        super(AbstractSite, self).__init__()

        # Computed metadata
        self.hash = None
        self.html = None
        self.method = 'GET'
        self.back_scrape_iterable = None
        self.downloader_executed = False
        self.cookies = {}
        self.cnt = cnt or CaseNameTweaker()
        self.request = {
            'verify': certifi.where(),
            'session': requests.session(),
            'headers': {'User-Agent': 'Juriscraper'},
            # Disable CDN caching on sites like SCOTUS (ahem)
            'cache-control': 'no-cache, no-store, max-age=1',
            'parameters': {},
            'request': None,
            'status': None,
            'url': None,
        }

        # Sub-classed metadata
        self.court_id = None
        self.url = None
        self.parameters = None
        self.uses_selenium = None
        self._opt_attrs = []
        self._req_attrs = []
        self._all_attrs = []

    def __str__(self):
        out = []
        for attr, val in self.__dict__.items():
            out.append('%s: %s' % (attr, val))
        return '\n'.join(out)

    def __iter__(self):
        for i in range(0, len(self.case_names)):
            yield self._make_item(i)

    def __getitem__(self, i):
        return self._make_item(i)

    def __len__(self):
        return len(self.case_names)

    def _make_item(self, i):
        """Using i, convert a single item into a dict. This is effectively a
        different view of the data.
        """
        item = {}
        for attr_name in self._all_attrs:
            attr_value = getattr(self, attr_name)
            if attr_value is not None:
                item[attr_name] = attr_value[i]
        return item

    def enable_test_mode(self):
        self.method = 'LOCAL'

    def disable_certificate_verification(self):
        """Scrapers that require this due to website misconfiguration
        should be checked periodically--calls to this method from
         site scrapers should be removed when no longer necessary.
         """
        self.request['verify'] = False

    def test_mode_enabled(self):
        return self.method == 'LOCAL'

    def to_json(self):
        return json.dumps(
            [item for item in self],
            default=json_date_handler,
        )

    def parse(self):
        if not self.downloader_executed:
            # Run the downloader if it hasn't been run already
            self.html = self._download()

            # Process the available html (optional)
            self._process_html()

        # Set the attribute to the return value from _get_foo()
        # e.g., this does self.case_names = _get_case_names()
        for attr in self._all_attrs:
            self.__setattr__(attr, getattr(self, '_get_%s' % attr)())

        self._clean_attributes()
        if 'case_name_shorts' in self._all_attrs:
            # This needs to be done *after* _clean_attributes() has been run.
            # The current architecture means this gets run twice. Once when we
            # iterate over _all_attrs, and again here. It's pretty cheap though.
            self.case_name_shorts = self._get_case_name_shorts()
        self._post_parse()
        self._check_sanity()
        self._date_sort()
        self._make_hash()
        return self

    def tweak_response_object(self):
        """
        Does nothing, but provides a hook that allows inheriting objects to
        tweak the requests object if necessary.
        """
        pass

    def _clean_text(self, text):
        """A hook for subclasses to override if needed."""
        return clean_html(text)

    def _clean_attributes(self):
        """Iterate over attribute values and clean them"""
        for attr in self._all_attrs:
            item = getattr(self, attr)
            if item is not None:
                cleaned_item = []
                for sub_item in item:
                    if attr == 'download_urls':
                        sub_item = sub_item.strip()
                    else:
                        if isinstance(sub_item, six.string_types):
                            sub_item = clean_string(sub_item)
                        elif isinstance(sub_item, datetime):
                            sub_item = sub_item.date()
                        if attr in ['case_names', 'docket_numbers']:
                            sub_item = harmonize(sub_item)
                    cleaned_item.append(sub_item)
                self.__setattr__(attr, cleaned_item)

    def _post_parse(self):
        """This provides an hook for subclasses to do custom work on the data
        after the parsing is complete.
        """
        pass

    def _check_sanity(self):
        """Check that the objects attributes make sense:
            1. Do all the attributes have the same length?
            1. Do we have any content at all?
            1. Is there a bare minimum of meta data?
            1. Are the dates datetime objects, not strings?
            1. Are any dates from the 22nd century? (01-01-2104)
            1. Are case_names more than just empty whitespace?
            1. Has the `cookies` attribute been normalized to a dict?
            1. ?

        The signature of this method is subject to change as additional checks
        become convenient.

        Inheriting classes should override this method calling super to give it
        the necessary parameters.

        If sanity is OK, no return value. If not, throw InsanityException or
        warnings, as appropriate.
        """
        lengths = {}
        for attr in self._all_attrs:
            if self.__getattribute__(attr) is not None:
                lengths[attr] = len(self.__getattribute__(attr))
        values = list(lengths.values())
        if values.count(values[0]) != len(values):
            # Are all elements equal?
            raise InsanityException("%s: Scraped meta data fields have differing"
                                    " lengths: %s" % (self.court_id, lengths))
        if len(self.case_names) == 0:
            logger.warning('%s: Returned with zero items.' % self.court_id)
        else:
            for field in self._req_attrs:
                if self.__getattribute__(field) is None:
                    raise InsanityException('%s: Required fields do not contain any data: %s' % (self.court_id, field))
            i = 0
            prior_case_name = None
            for name in self.case_names:
                if not name.strip():
                    raise InsanityException(
                        "Item with index %s has an empty case name. The prior "
                        "item had case name of: %s" % (i, prior_case_name)
                    )
                prior_case_name = name
                i += 1

        for index, case_date in enumerate(self.case_dates):
            if not isinstance(case_date, date):
                raise InsanityException(
                    '%s: member of case_dates list not a valid date object. '
                    'Instead it is: %s with value: %s' % (
                        self.court_id, type(case_date), case_date)
                )
            # Sanitize case date, fix typo of current year if present
            fixed_date = fix_future_year_typo(case_date)
            if fixed_date != case_date:
                logger.info(
                    "Date year typo detected. Converting %s to %s "
                    "for case '%s' in %s" % (case_date, fixed_date, self.case_names[index], self.court_id)
                )
                case_date = fixed_date
                self.case_dates[index] = fixed_date
            if case_date.year > 2025:
                raise InsanityException(
                    '%s: member of case_dates list is from way in the future, '
                    'with value %s' % (self.court_id, case_date.year)
                )

        # Is cookies a dict?
        if type(self.cookies) != dict:
            raise InsanityException('self.cookies not set to be a dict by '
                                    'scraper.')
        logger.info("%s: Successfully found %s items." % (self.court_id,
                                                          len(self.case_names)))

    def _date_sort(self):
        """ Sort the object by date.
        """
        if len(self.case_names) > 0:
            obj_list_attrs = [self.__getattribute__(attr) for attr in
                              self._all_attrs if
                              isinstance(self.__getattribute__(attr), list)]
            zipped = list(zip(*obj_list_attrs))
            zipped.sort(reverse=True)
            i = 0
            obj_list_attrs = list(zip(*zipped))
            for attr in self._all_attrs:
                if isinstance(self.__getattribute__(attr), list):
                    self.__setattr__(attr, obj_list_attrs[i][:])
                    i += 1

    def _make_hash(self):
        """Make a unique ID. ETag and Last-Modified from courts cannot be
        trusted
        """
        self.hash = hashlib.sha1(str(self.case_names).encode()).hexdigest()

    def _get_adapter_instance(self):
        """Hook for returning a custom HTTPAdapter

        This function allows subclasses to do things like explicitly set
        specific SSL configurations when being called. Certain courts don't work
        unless you specify older versions of SSL.
        """
        return HTTPAdapter()

    def _make_html_tree(self, text):
        """Hook for custom HTML parsers

        By default, the etree.html parser is used, but this allows support for
        other parsers like the html5parser or even BeautifulSoup, if it's called
        for (example: return get_html5_parsed_text(text)). Otherwise, this method
        can be overwritten to execute custom parsing logic.
        """
        return get_html_parsed_text(text)

    def _download(self, request_dict={}):
        """Download the latest version of Site"""
        self.downloader_executed = True
        if self.method == 'POST':
            truncated_params = {}
            for k, v in self.parameters.items():
                truncated_params[k] = trunc(v, 50, ellipsis='...[truncated]')
            logger.info("Now downloading case page at: %s (params: %s)" % (self.url, truncated_params))
        else:
            logger.info("Now downloading case page at: %s" % self.url)
        self._process_request_parameters(request_dict)
        if self.method == 'GET':
            self._request_url_get(self.url)
        elif self.method == 'POST':
            self._request_url_post(self.url)
        elif self.test_mode_enabled():
            self._request_url_mock(self.url)
        self._post_process_response()
        return self._return_response_text_object()

    def _process_html(self):
        """Hook for processing available self.html after it's been downloaded.
        This step is completely optional, but is useful if you want to transform
        the html before running the data getters (_get_*), or if its easier to
        extract the cases form the html in a linear function (here) as opposed
        to using separate data getters.  See sc.py for example.
        """
        pass

    def _process_request_parameters(self, parameters={}):
        """Hook for processing injected parameter overrides"""
        if parameters.get('verify') is not None:
            self.request['verify'] = parameters['verify']
            del parameters['verify']
        self.request['parameters'] = parameters
        self.request['session'].mount('https://', self._get_adapter_instance())

    def _request_url_get(self, url):
        """Execute GET request and assign appropriate request dictionary
        values
        """
        self.request['url'] = url
        self.request['response'] = self.request['session'].get(
            url,
            headers=self.request['headers'],
            verify=self.request['verify'],
            timeout=60,
            **self.request['parameters']
        )

    def _request_url_post(self, url):
        """Execute POST request and assign appropriate request dictionary values"""
        self.request['url'] = url
        self.request['response'] = self.request['session'].post(
            url,
            headers=self.request['headers'],
            verify=self.request['verify'],
            data=self.parameters,
            timeout=60,
            **self.request['parameters']
        )

    def _request_url_mock(self, url):
        """Execute mock request, used for testing"""
        self.request['url'] = url
        self.request['response'] = MockRequest(url=self.url).get()

    def _post_process_response(self):
        """Cleanup to response object"""
        self.tweak_response_object()
        self.request['response'].raise_for_status()
        set_response_encoding(self.request['response'])

    def _return_response_text_object(self):
        if self.request['response']:
            if 'json' in self.request['response'].headers.get('content-type', ''):
                return self.request['response'].json()
            else:
                if six.PY2:
                    payload = self.request['response'].text
                else:
                    payload = str(self.request['response'].content)

                text = self._clean_text(payload)
                html_tree = self._make_html_tree(text)
                html_tree.rewrite_links(fix_links_in_lxml_tree,
                                        base_href=self.request['url'])
                return html_tree

    def _get_html_tree_by_url(self, url, parameters={}):
        self._process_request_parameters(parameters)
        self._request_url_get(url)
        self._post_process_response()
        tree = self._return_response_text_object()
        tree.make_links_absolute(url)
        return tree

    def _download_backwards(self):
        # methods for downloading the entire Site
        pass

    @staticmethod
    def cleanup_content(content):
        """
          Given the HTML from a page, the binary PDF file, or similar, do any
          last-minute cleaning.

          This method should be called as the last step by any caller and works
          to do any cleanup that is necessary. Usually, this is needed on HTML
          pages, in jurisdictions that post their content in an HTML page with
          headers, footers and other content must be stripped after the page
          has been downloaded by the caller.
        """
        return content

    def _get_cookies(self):
        """
          Some websites require cookies in order to be scraped. This method
          provides a hook where cookies can be retrieved by calling functions.
          Generally the cookies will be set by the _download() method.

          self.cookies is a list of dicts of the form:
          [
              {
                  u'domain':   u'www.search.txcourts.gov',
                  u'httponly': True,
                  u'name':     u'ASP.NET_SessionId',
                  u'path':     u'/',
                  u'secure':   False,
                  u'value':    u'hryku05534xhgr45yxvusuux'
              },
          ]
        """
        return self._cookies

    def _get_case_name_shorts(self):
        """Generates short case names for all the case names that we scrape."""
        case_name_shorts = []
        for case_name in self.case_names:
            case_name_shorts.append(self.cnt.make_case_name_short(case_name))
        return case_name_shorts

    def _get_blocked_statuses(self):
        """Should these items be blocked by search engines? Default is False for
        all subclasses, indicating that the items should not be blocked.

        This method is important because some courts (like family or asylum
        courts) should choose privacy over openness. Note that we consider
        these functions to be a hint to callers, so following these guidelines
        is not guaranteed.
        """
        return [False] * len(self.case_names)



# -*- coding: utf-8 -*-
import os
import sys
import warnings

from requests.exceptions import ConnectionError
from requests.models import Request, Response

from .exceptions import SlownessException

WARN_SLOW_SRAPERS = 'TRAVIS' in os.environ
warnings.simplefilter("always")


class SlownessWarning(UserWarning):
    pass


class CompareFileGeneratedWarning(UserWarning):
    pass


class TooSlowWarning(SlownessWarning):
    pass


class MockRequest(Request):
    def __init__(self, url=None):
        super(Request, self).__init__()
        self.url = url
        #: Resulting :class:`HTTPError` of request, if one occurred.
        self.error = None

        #: Encoding to decode with when accessing r.content.
        self.encoding = None

        #: The :class:`Request <Request>` that created the Response.
        self.request = self

    def get(self):
        r = Response()
        try:
            with open(self.url, mode='rb') as stream:
                r._content = stream.read()
                #: Integer Code of responded HTTP Status.
                r.status_code = 200
                if self.url.endswith('json'):
                    r.headers['content-type'] = 'application/json'
        except IOError as e:
            r.status_code = 404
            raise ConnectionError(e)

        r._content_consumed = True

        #: Final URL location of Response.
        r.url = self.url

        # Return the response.
        return r


def warn_or_crash_slow_parser(duration, warn_duration=1, max_duration=15):
    msg = ''
    if duration > max_duration:
        if sys.gettrace() is None and not WARN_SLOW_SRAPERS:
            # Only do this if we're not debugging. Debuggers make things slower
            # and breakpoints make things stop.
            raise SlownessException(
                "This scraper took {duration}s to test, which is more than "
                "the allowed speed of {max_duration}s. Please speed it up for "
                "tests to pass.".format(duration=duration,
                                        max_duration=max_duration)
            )
        elif WARN_SLOW_SRAPERS:
            msg = ' - WARNING: SLOW SCRAPER'
            wmsg = "WARNING: VERY SLOW SCRAPER (potential Error)"
            warnings.warn(wmsg, TooSlowWarning)
    elif duration > warn_duration:
        msg = ' - WARNING: SLOW SCRAPER'
        warnings.warn('WARNING: SLOW SCRAPER', SlownessWarning)
    else:
        msg = ''
    return msg


def warn_generated_compare_file(path_to_compare_file):
    warning = 'WARNING: GENERATED COMPARE FILE: %s' % path_to_compare_file
    warnings.warn(warning, CompareFileGeneratedWarning)



"""Search for cases and parse the iquery.pl search results.

This lets you look up cases and parse results. Note that the classes here are
similar but distinct from CaseQuery, which is used with a particular case and
returns a menu of options as a result.

In the cases, we look up search results by name, date, etc.
"""
import pprint
import sys

from .docket_report import BaseDocketReport
from .reports import BaseReport
from .utils import get_pacer_case_id_from_nonce_url
from ..lib.log_tools import make_default_logger
from ..lib.string_utils import clean_string, convert_date_string, \
    force_unicode, harmonize
from ..lib.utils import clean_court_object

logger = make_default_logger()


class BaseCaseQueryAdvanced(BaseDocketReport, BaseReport):
    """Base query for both district and bankruptcy queries.
    """
    PATH = 'cgi-bin/iquery.pl'

    CACHE_ATTRS = ['metadata']

    def __init__(self, court_id, pacer_session=None):
        super(BaseCaseQueryAdvanced, self).__init__(court_id, pacer_session)
        # Initialize the empty cache properties
        self._clear_caches()
        self._metadata = None

    def parse(self):
        self._clear_caches()
        super(BaseCaseQueryAdvanced, self).parse()

    @property
    def data(self):
        """Get all the data back from this endpoint.

        Don't attempt to return parties or docket_entries like the superclass
        does.
        """
        if self.is_valid is False:
            return {}

        return self.metadata

    @staticmethod
    def get_text_for_cell(cell):
        """Return the text within a <td>"""
        return force_unicode(cell.text_content().strip())

    def get_date_for_cell(self, cell):
        """Return a parsed date within a <td>"""
        s = self.get_text_for_cell(cell)
        try:
            return convert_date_string(s)
        except ValueError:
            # Happens when the value isn't a date, like when it's "N/A", say.
            return None


class CaseQueryAdvancedBankruptcy(BaseCaseQueryAdvanced):
    """iQuery.pl for bankruptcy cases"""

    @property
    def metadata(self):
        if self._metadata is not None:
            return self._metadata

        # The data we're after look like this (re-spacing):
        #
        # <TR class=rowBackground1>
        #     <TD VALIGN=center>PacifiCorp (pty) <br> (1 case)</TD>
        #     <TD NOWRAP VALIGN=center><A
        #             HREF="iquery.pl?440916540563012-L_1_0-0-12612">01-30923</A>
        #     </TD>
        #     <TD VALIGN=center>Pacific Gas and Electric Co.</TD>
        #     <TD VALIGN=center> 11</TD>
        #     <TD VALIGN=center> 04/06/01</TD>
        #     <TD VALIGN=center> Creditor</TD>
        #     <TD VALIGN=center>N / A</TD>
        # </TR>
        #
        # With the following columns (in bankruptcy):
        #    1. Name  (we skip this for now)
        #    2. Case Number
        #    3. Case Title
        #    4. Chapter / Lead BK case
        #    5. Date Filed
        #    6. Party Role (we skip this row for now)
        #    7. Date Closed
        #
        # Columns 1 and 6 are not always present, and so as a first pass below
        # we just nuke them. They're not that important to us now anyway. But
        # none would deny this is a hackish way to normalize the results.

        table_rows = self.tree.xpath('//table//tr[@class="rowBackground1" or '
                                     '@class="rowbackground2"]')
        data = []

        for table_row in table_rows:
            cells = table_row.xpath('./td')
            if len(cells) == 7:
                # There are person and case results. Eliminate the first and
                # sixth cells to normalize with tables that lack person results
                # This is a hack that makes debugging harder. Better would be
                # to sniff the header row.
                del cells[5]
                del cells[0]

            row_data = {
                'docket_number': self.get_text_for_cell(cells[0]),
                'case_name': clean_string(
                    harmonize(self.get_text_for_cell(cells[1]))),
                'chapter': self.get_text_for_cell(cells[2]),
                'date_filed': self.get_date_for_cell(cells[3]),
                'date_closed': self.get_date_for_cell(cells[4]),
            }
            href = cells[0].xpath('.//@href')[0]
            row_data['pacer_case_id'] = get_pacer_case_id_from_nonce_url(href)

            data.append(row_data)

        data = clean_court_object(data)

        self._metadata = data
        return data

    def query(self, name_last='', name_first='', name_middle='',
              person_type='', filed_from=None, filed_to=None,
              last_entry_from=None, last_entry_to=None, natures_of_suit=''):
        """Use a bankruptcy court's PACER query function to look up cases

        At the top of every district PACER court, there's a button that says,
        "Query". When you click that button, it either does a query for the
        criteria you specify, returning a list of results or it returns the
        specific case you were looking for, including metadata and a list of
        reports you can run on that case.

        This method only supports the use case where you are looking up
        criteria. For the other use case, please see the CaseQuery object.
        SSN and Tax ID filtering not yet implemented.

        :param name_last: The last name to look up.
        :param name_first: The first name to look up.
        :param name_middle: The middle name to look up.
        :param person_type: What type the person is. See PACER for valid
        entries.
        :param filed_from: The date filed entry for the earliest possible
        filing (inclusive).
        :type filed_from: datetime.date
        :param filed_to: The date filed value for the latest possible filing
        (inclusive).
        :type filed_to: datetime.date
        :param last_entry_from: The earliest date that the last entry on the
        docket can be from. I think.
        :type last_entry_from: datetime.date
        :param last_entry_to: The latest date that the last entry on the docket
        can be from. I think.
        :type last_entry_to: datetime.date
        :param natures_of_suit: A Python list of nature of suit numbers that to
        use for filtering. (See PACER for values.)

        :return None: Instead, sets self.response attribute and runs
        self.parse()
        """
        assert self.session is not None, \
            "session attribute of DocketReport cannot be None."
        assert all([filed_from, filed_to]) or \
            not any([filed_from, filed_to]), \
            "Both or neither of filing date fields must be complete."
        assert all([last_entry_from, last_entry_to]) or \
            not any([last_entry_from, last_entry_to]), \
            "Both or neither of last entry date fields must be complete."

        # PACER only allows so many days per query. Exceed this threshold and
        # you get an error message and zero results.
        max_days = 31
        max_days_msg = "PACER has a %s day limit on date filters. " \
                       "Narrow your search." % max_days
        if filed_from:
            assert (filed_to - filed_from).days + 1 < max_days, \
                max_days_msg
        if last_entry_from:
            assert (last_entry_to - last_entry_from).days + 1 < max_days, \
                max_days_msg

        params = {}
        if filed_from:
            params[u'filed_from'] = filed_from.strftime(u'%m/%d/%Y')
        if filed_to:
            params[u'filed_to'] = filed_to.strftime(u'%m/%d/%Y')
        if last_entry_from:
            params[u'lastentry_from'] = last_entry_from.strftime(u'%m/%d/%Y')
        if last_entry_to:
            params[u'lastentry_to'] = last_entry_to.strftime(u'%m/%d/%Y')

        params.update({
            # Name fields
            u'last_name': name_last,
            u'first_name': name_first,
            u'middle_name': name_middle,
            u'person_type': person_type,
        })
        logger.info(u"Running advanced case query with params '%s'", params)
        self.response = self.session.post(self.url + '?1-L_1_0-1', data=params)
        self.parse()


class CaseQueryAdvancedDistrict(BaseCaseQueryAdvanced):
    """iQuery.pl for district cases

    Not yet implemented. Unfortunately both the query parameters needed for the
    query() method and the results parsed by the metadata() method are unique
    across district and bankruptcy.
    """
    def __init__(self):
        raise NotImplementedError("This object is a stub.")

    def query(self, *args, **kwargs):
        raise NotImplementedError("This object is a stub")


def _main():
    if len(sys.argv) != 2:
        print("Usage: python -m juriscraper.pacer.case_query_advanced filepath")
        print("Please provide a path to an HTML file to parse.")
        sys.exit(1)

    # Court ID is only needed for querying. Actual
    # parsed value appears in output
    report = CaseQueryAdvancedBankruptcy('mad')
    filepath = sys.argv[1]
    print("Parsing HTML file at %s" % filepath)
    with open(filepath, 'r') as f:
        text = f.read().decode('utf-8')
    report._parse_text(text)
    pprint.pprint(report.data, indent=2)


if __name__ == "__main__":
    _main()



#  Scraper for Fourth Circuit of Appeals
# CourtID: ca4
# Court Short Name: ca4
# Author: Andrei Chelaru
# Reviewer: mlr
# Date created: 18 July 2014

from juriscraper.OralArgumentSite import OralArgumentSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OralArgumentSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.ca4.uscourts.gov/oral-argument/listen-to-oral-arguments'

    def _get_download_urls(self):
        return [href for href in self.html.xpath('//tr/td[2]//a/@href')]

    def _get_case_names(self):
        return self.text_from_cell(3)

    def _get_case_dates(self):
        return [convert_date_string(date) for date in self.text_from_cell(1)]

    def _get_judges(self):
        return self.text_from_cell(4)

    def _get_docket_numbers(self):
        return self.text_from_cell(2)

    def text_from_cell(self, cell_number):
        return [cell.text_content().strip() for cell in self.html.xpath('//tr/td[%d]' % cell_number)]



# Scraper for California's Second District Court of Appeal
# CourtID: calctapp_2nd
# Court Short Name: Cal. Ct. App.

from juriscraper.opinions.united_states.state import calctapp_1st


class Site(calctapp_1st.Site):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court_code = 'B'
        self.division = '2nd App. Dist.'



# Scraper for Iowa Supreme Court
# CourtID: iowa
# Court Short Name: iowa

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.cases = []
        self.archive = False
        self.back_scrape_iterable = ['placeholder']  # this array can't be empty
        self.url = 'https://www.iowacourts.gov/iowa-courts/supreme-court/supreme-court-opinions/'

    def _download(self, request_dict={}):
        html = super(Site, self)._download(request_dict)
        self.extract_cases(html)
        if self.test_mode_enabled() or self.archive:
            return html

        # Walk over pagination "Next" page(s), if present
        proceed = True
        while proceed:
            next_page_url = self.extract_next_page_url(html)
            if next_page_url:
                logger.info('Scraping next page: %s' % next_page_url)
                html = self._get_html_tree_by_url(next_page_url)
                self.extract_cases(html)
            else:
                proceed = False

    def _get_case_names(self):
        return [case['name'] for case in self.cases]

    def _get_download_urls(self):
        return [case['url'] for case in self.cases]

    def _get_case_dates(self):
        return [convert_date_string(case['date']) for case in self.cases]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_dates)

    def _get_docket_numbers(self):
        return [case['docket'] for case in self.cases]

    def _download_backwards(self, _):
        """Walk over all "Archive" links on Archive page,
        extract cases dictionaries, and add to self.cases
        """
        self.archive = True
        self.url = self.url + 'opinions-archive/'
        landing_page_html = self._download()
        path = '//div[@class="main-content-wrapper"]//a[contains(./text(), "Opinions Archive")]/@href'
        for archive_page_url in landing_page_html.xpath(path):
            logger.info("Back scraping archive page: %s" % archive_page_url)
            archive_page_html = self._get_html_tree_by_url(archive_page_url)
            self.extract_archive_cases(archive_page_html)

    def extract_cases(self, html):
        """Extract case dictionaries from "Recent" html page
        and add them to self.cases
        """
        case_substring = 'Case No.'
        case_elements = html.xpath('//h3[contains(., "%s")]' % case_substring)
        for case_element in case_elements:
            text = case_element.text_content()
            parts = text.split(':')
            docket = parts[0].replace(case_substring, '').strip()
            name = parts[1].strip()
            date_text = case_element.xpath('./following::p[1]')[0].text_content()
            date_string = date_text.replace('Filed', '')
            url = case_element.xpath('./following::p[2]//a/@href')[0]
            self.cases.append({
                'name': name,
                'docket': docket,
                'date': date_string,
                'url': url,
            })

    def extract_archive_cases(self, html):
        """Extract case dictionaries from "Archive" html page
        and add them to self.cases
        """
        path_date = '//div[@class="cms_category_icon_title_row"]'
        for date_header in html.xpath(path_date):
            text = date_header.text_content()
            date_string = text.replace('- DELETE', '')
            path_cases = './following::div[@class="cms_items"][1]/div[@class="cms_item"]'
            for case_container in date_header.xpath(path_cases):
                docket_element = case_container.xpath('./div[@class="cms_item_icon_title_row"]')[0]
                self.cases.append({
                    'date': date_string,
                    'url': docket_element.xpath('.//a/@href')[0],
                    'docket': docket_element.text_content().strip(),
                    'name': case_container.xpath('./div[@class="cms_item_description"]')[0].text_content().strip(),
                })

    def extract_next_page_url(self, html):
        """Return the href url from "Next" pagination element
        if it exists, otherwise return False.
        """
        path = '//div[contains(./@class, "pagination-next-page")]//a/@href'
        elements = html.xpath(path)
        return elements[0] if elements else False



# -*- coding: utf-8 -*-

"""
Contact: Sara Velasquez, svelasquez@idcourts.net, 208-947-7501
History:
 - 2014-08-05, mlr: Updated.
 - 2015-06-19, mlr: Updated to simply the XPath expressions and to fix an OB1
   problem that was causing an InsanityError. The cause was nasty HTML in their
   page.
 - 2015-10-20, mlr: Updated due to new page in use.
 - 2015-10-23, mlr: Updated to handle annoying situation.
 - 2016-02-25 arderyp: Updated to catch "ORDER" (in addition to "Order") in download url text
"""
import six
from lxml import html

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string, clean_if_py3


class Site(OpinionSite):
    # Skip first row of table, it's a header
    starting_table_row = '//table//tr[position() > 1]'
    # Skip rows that don't have  link in 4th cell with
    # either 'Opinion', 'Order', 'ORDER', or 'Amend' in
    # the link text
    row_path_conditional = '/td[4]//a[' \
        'contains(.//text(), "Opinion") or ' \
        'contains(.//text(), "Order") or ' \
        'contains(.//text(), "ORDER") or ' \
        'contains(.//text(), "Amended")]'
    base_path = '%s[./%s]' % (starting_table_row, row_path_conditional)

    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.url = 'https://www.isc.idaho.gov/appeals-court/isc_civil'
        self.court_id = self.__module__

    def _get_case_names(self):
        case_names = []
        path = '%s/td[3]' % self.base_path
        for cell in self.html.xpath(path):
            name_string = html.tostring(cell, method='text', encoding='unicode')
            name_string = clean_if_py3(name_string).strip()
            if name_string:
                case_names.append(name_string)
        return case_names

    def _get_download_urls(self):
        # We'll accept an order document if the opinion document
        # is missing. But we obviously prefer the opinion doc,
        # so we put it first in the order below, and we take the
        # first link that matches in path in each 4th cell.
        path = '%s%s[1]/@href' % (self.starting_table_row, self.row_path_conditional)
        return [url for url in self.html.xpath(path)]

    def _get_case_dates(self):
        case_dates = []
        path = '%s/td[1]' % self.base_path
        for cell in self.html.xpath(path):
            date_string = html.tostring(cell, method='text', encoding='unicode')
            date_string = clean_if_py3(date_string).strip()
            if date_string:
                if six.PY2:
                    date_string = date_string.encode('ascii', 'ignore')
                date_string = date_string.replace('Sept ', 'Sep ')  # GIGO!  (+1 by arderyp)
                case_dates.append(convert_date_string(date_string))
        return case_dates

    def _get_docket_numbers(self):
        path = '%s/td[2]//text()' % self.base_path
        return [text.strip() for text in self.html.xpath(path) if text.strip()]

    def _get_precedential_statuses(self):
        return ["Published"] * len(self.case_names)



# -*- coding: utf-8 -*-

# Scraper for Pennsylvania Supreme Court
# CourtID: pa
# Court Short Name: pa

import re

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string
from juriscraper.lib.string_utils import clean_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.regex = False
        self.url = 'http://www.pacourts.us/assets/rss/SupremeOpinionsRss.ashx'
        self.set_regex(r"(.*)(?:[,-]?\s+Nos?\.)(.*)")
        self.base = "//item[not(contains(title/text(), 'Judgment List'))]" \
                    "[not(contains(title/text(), 'Reargument Table'))]" \
                    "[contains(title/text(), 'No.')]"
        self.cases = []

    def _download(self, request_dict={}):
        html = super(Site, self)._download(request_dict)
        self._extract_case_data_from_html(html)
        return html

    def _extract_case_data_from_html(self, html):
        for item in html.xpath(self.base):
            creator = item.xpath('./creator')[0].text_content()
            pubdate = item.xpath('./pubdate')[0].text_content()
            pubdate_sanitized = self.sanitize_text(pubdate)
            title = item.xpath('./title')[0].text_content()
            title_sanitized = self.sanitize_text(title)
            title_clean = clean_string(title_sanitized)
            search = self.regex.search(title_clean)
            url = item.xpath('.//@href')[0]

            if search:
                name = search.group(1)
                docket = search.group(2)
            else:
                name = title_clean
                docket = self._extract_docket_from_url(url)

            self.cases.append({
                'name': name,
                'date': convert_date_string(pubdate_sanitized),
                'docket': docket,
                'judge': self.sanitize_text(creator),
                'url': url,
            })

    def _extract_docket_from_url(self, url):
        """Sometimes the court doesnt include the docket number in the title,
        in which case we need to derive it based on the opinion url (it is
        included in the PDF file name)
        """
        parts = url.split('/')[-1].split('CD')
        number = parts[0]
        year_suffix_text = parts[1].split('_')[0]
        year_suffix = re.sub("[^0-9]", "", year_suffix_text)  # Strip non-numeric characters
        return '%s C.D. 20%s' % (number, year_suffix)

    def _get_case_names(self):
        return [case['name'] for case in self.cases]

    def _get_download_urls(self):
        return [case['url'] for case in self.cases]

    def _get_case_dates(self):
        return [case['date'] for case in self.cases]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.cases)

    def _get_docket_numbers(self):
        return [case['docket'] for case in self.cases]

    def _get_judges(self):
        return [case['judge'] for case in self.cases]

    def sanitize_text(self, text):
        text = clean_string(text)
        return text.replace(r'\n', '\n').replace(u'–', '-')

    def set_regex(self, pattern):
        self.regex = re.compile(pattern)



"""Scraper for Missouri
CourtID: mo
Court Short Name: MO
Author: Ben Cassedy
Date created: 04/27/2014
"""

from datetime import date

from juriscraper.OpinionSite import OpinionSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url_slug = "Supreme"
        self.url = self.build_url()
        self.cases = []

    def build_url(self):
        url_template = 'https://www.courts.mo.gov/page.jsp?id=12086&dist=Opinions %s&date=all&year=%s#all'
        return url_template % (self.url_slug, date.today().year)

    def _download(self, request_dict={}):
        html = super(Site, self)._download(request_dict)
        self._extract_cases_from_html(html)
        return html

    def _extract_cases_from_html(self, html):
        for date_block in html.xpath("//div[@id='content']/div/form/table/tr/td"):
            date_string = date_block.xpath('input/@value')
            if date_string:
                for case_block in date_block.xpath('table/tr/td'):
                    links = case_block.xpath('a')
                    first_link_text = links[0].xpath('text()')
                    if first_link_text and 'Orders Pursuant to Rules' in first_link_text[0]:
                        continue

                    target_link_index = 1 if len(links) > 1 else 0
                    bolded_text = case_block.xpath('b')
                    docket = self.sanitize_docket(bolded_text[0].xpath('text()')[0])
                    text = case_block.xpath('text()')
                    judge, disposition = self.parse_judge_disposition_from_text(text)

                    self.cases.append({
                        'date': convert_date_string(date_string[0]),
                        'docket': docket,
                        'judge': judge,
                        'url': links[target_link_index].xpath('@href')[0],
                        'name': links[target_link_index].xpath('text()')[0],
                        'disposition': disposition,
                    })

    @staticmethod
    def sanitize_docket(docket):
        for substring in [':', 'and', '_', 'Consolidated', '(', ')', ',']:
            docket = docket.replace(substring, ' ')
        return ', '.join(docket.split())

    @staticmethod
    def parse_judge_disposition_from_text(text_raw_list):
        text_clean_list = [text.strip() for text in text_raw_list if text.strip()]
        return text_clean_list[0], text_clean_list[1]

    def _get_download_urls(self):
        return [case['url'] for case in self.cases]

    def _get_case_names(self):
        return [case['name'] for case in self.cases]

    def _get_case_dates(self):
        return [case['date'] for case in self.cases]

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.cases)

    def _get_docket_numbers(self):
        return [case['docket'] for case in self.cases]

    def _get_judges(self):
        return [case['judge'] for case in self.cases]

    def _get_dispositions(self):
        return [case['disposition'] for case in self.cases]



"""Oral Argument Audio Scraper for Court of Appeals for the Sixth Circuit
CourtID: ca6
Court Short Name: 6th Cir.
Court Contact: WebSupport@ca6.uscourts.gov
Authors: Brian W. Carver, Michael Lissner
Reviewer: None
History:
  2014-11-06: Started by Brian W. Carver and wrapped up by mlr.
  2016-06-30: Updated by mlr.
"""

import re
from datetime import datetime
from six.moves.urllib.parse import urlparse, urljoin, parse_qs

from juriscraper.OralArgumentSite import OralArgumentSite
from juriscraper.lib.string_utils import convert_date_string


class Site(OralArgumentSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.opn.ca6.uscourts.gov/internet/court_audio/aud1.php'
        self.xpath_root = '//table[@class="views-table cols-3"]'
        self.regex = re.compile('((?:\d{2}[- ]\d{4}\s*)+)(.*)')
        self.back_scrape_iterable = ['nothing']  # Just a placeholder for this court
        self.backscrape = False

    def _get_download_urls(self):
        """Two options are currently provided by the site. The first is a link
        to "save" the file, which gives you a zip containing the file. The
        second is a link to "play" the file, which takes you to a flash player.

        The good news is that the link to "play" it contains a real link to
        actually download it inside the 'link' param.
        """
        if self.backscrape:
            path_to_flash_page = '//tr/td[3]/a/@href[contains(., "?link=")]'
        else:
            path_to_flash_page = '//tr/td[2]/a/@href[contains(., "?link=")]'
        links_to_flash = list(self.html.xpath(path_to_flash_page))
        urls = []
        for url in links_to_flash:
            path = parse_qs(urlparse(url).query)['link'][0]
            # Remove newlines and line returns from urls.
            path = path.replace('\n', '').replace('\r', '')

            if 'www.opn' not in url:
                # Update the URL if it's not the one we want.
                url = url.replace('www', 'www.opn')
            urls.append(urljoin(url, path))
        return urls

    def _get_case_names(self):
        if self.backscrape:
            path = self.xpath_root + '//td[2]/text()'
        else:
            path = self.xpath_root + '/tr/td[1]/text()'
        case_names = []
        for s in self.html.xpath(path):
            case_names.append(self.regex.search(s).group(2))
        return case_names

    def _get_case_dates(self):
        dates = []
        if self.backscrape:
            date_strs = self.html.xpath('//table//td[1]//text()')
            return [convert_date_string(s) for s in date_strs]
        else:
            # Multiple items are listed under a single date.
            date_path = './/th[1]'
            # For every table full of OA's...
            for table in self.html.xpath(self.xpath_root):
                # Find the date str, e.g. "10-10-2014 - Friday"
                date_str = table.xpath(date_path)[0].text_content()
                d = datetime.strptime(date_str[:10], '%m-%d-%Y').date()

                # The count of OAs on a date is the number of rows minus the
                # header row.
                total_rows = len(table.xpath('.//tr')) - 1
                dates.extend([d] * total_rows)
            return dates

    def _get_docket_numbers(self):
        if self.backscrape:
            path = self.xpath_root + '//td[2]/text()'
        else:
            path = self.xpath_root + '/tr/td[1]/text()'
        return [self.regex.search(s).group(1).strip().replace(' ', '-') for
                s in self.html.xpath(path)]

    def _download_backwards(self, _):
        """You can get everything with a single POST, thus we just ignore the
        back_scrape_iterable.
        """
        self.backscrape = True
        self.method = 'POST'
        self.xpath_root = '//table'
        self.url = 'http://www.opn.ca6.uscourts.gov/internet/court_audio/audSearchRes.php'
        self.parameters = {
            'caseNumber': "",
            'shortTitle': "",
            'dateFrom': '01/01/2013',
            'dateTo': '01/01/2015',
            'Submit': 'Submit+Query',
        }
        self.html = self._download()



"""Scraper for Second Circuit
CourtID: ca2
Author: MLR
Reviewer: MLR
Contact:
  Webmaster@ca2.uscourts.gov ('Calendar Team' handles updates, and is responsive)
  ecfhelpdesk@ca2.uscourts.gov
  Shane_Clouden@ca2.uscourts.gov
History:
  2016-09-09: Created by MLR
"""

from lxml import html

from juriscraper.OralArgumentSite import OralArgumentSite
from juriscraper.lib.html_utils import get_html5_parsed_text
from juriscraper.lib.string_utils import convert_date_string


class Site(OralArgumentSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.ca2.uscourts.gov/decisions'
        self.method = 'POST'
        self.parameters = {
            'IW_SORT': '-DATE',
            'IW_BATCHSIZE': '100',
            'IW_FIELD_TEXT': '*',
            'IW_DATABASE': 'Oral Args'
        }
        self.uses_selenium = False
        self.base_xpath = '//tr[contains(.//a/@href, "mp3")]'

    def _make_html_tree(self, text):
        return get_html5_parsed_text(text)

    def _get_download_urls(self):
        path = '//@href[contains(., "mp3")]'
        return list(self.html.xpath(path))

    def _get_case_names(self):
        case_names = []
        for e in self.html.xpath('%s/td[2]' % self.base_xpath):
            s = html.tostring(e, method='text', encoding='unicode')
            case_names.append(s)
        return case_names

    def _get_case_dates(self):
        path = '%s/td[3]' % self.base_xpath
        dates = []
        for e in self.html.xpath(path):
            date_string = html.tostring(e, method='text', encoding='unicode')
            # For some reason, some records have mysterious
            # alpha characters after the date, in the date
            # column (example: '9-28-16 B'). Strip it here.
            date_string = date_string.strip().split()[0]
            dates.append(convert_date_string(date_string))
        return dates

    def _get_docket_numbers(self):
        path = '%s/td[1]//text()' % self.base_xpath
        return list(self.html.xpath(path))



"""Scraper for Eleventh Circuit of Appeals
CourtID: ca11
Court Short Name: ca11
Author: Jon Andersen
Reviewer: mlr
Date created: 28 Aug 2018
"""

from datetime import datetime

from juriscraper.OralArgumentSite import OralArgumentSite


class Site(OralArgumentSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.ca11.uscourts.gov/oral-argument-recordings'
        self.back_scrape_iterable = [i for i in range(0, 52)]
        self.base_path = ("//tr[contains(@class, 'odd') or "
                          "contains(@class, 'even')]")

    def _get_download_urls(self):
        path = self.base_path + "//td[5]//@href"
        return list(self.html.xpath(path))

    def _get_case_names(self):
        path = self.base_path + "//td[2]/text()"
        return list(self.html.xpath(path))

    def _get_case_dates(self):
        path = self.base_path + "//td[3]/span/text()"
        return map(self._return_case_date, self.html.xpath(path))

    @staticmethod
    def _return_case_date(s):
        return datetime.strptime(s.strip(), '%Y-%m-%d').date()

    def _get_docket_numbers(self):
        path = self.base_path + "//td[1]/text()"
        # normalize docket numbers
        # get rid of "consolidated with" text
        # parse docket numbers like docketnum1 & docketnum2
        # also handle docketnum1\ndocketnum2\ndocketnum3
        # Return comma joined string like docketnum1, docketnum2
        return [", ".join(d.strip().upper()
                .replace("&", "\n")
                .replace(" AND ", "\n")
                .replace(",", "\n")
                .replace("CONS. WITH", "\n")
                .replace("CONSOLIDATED WITH", "\n")
                .split())
            for d in self.html.xpath(path)]

    def _download_backwards(self, i):
        self.url = 'http://www.ca11.uscourts.gov/oral-argument-recordings?page={i}'.format(
            i=i,
        )
        self.html = self._download()



# Scraper for Louisiana Supreme Court
# CourtID: la
# Court Short Name: LA
# Author: Andrei Chelaru
# Reviewer: mlr
# Date: 2014-07-05

import re
import time
from lxml import html
from datetime import date

from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSite import OpinionSite


class Site(OpinionSite):
    def __init__(self, *args, **kwargs):
        super(Site, self).__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = 'http://www.lasc.org/news_releases/{year}/default.asp'.format(year=date.today().year)
        self.path_case_paragraph = (
            "//p[contains(., 'v.') or "
            "contains(., 'IN RE') or "
            "contains(., 'IN THE') or "
            "contains(., 'vs.')]"
        )
        self.path_case_link = '%s//a' % self.path_case_paragraph

    def _download(self, request_dict={}):
        html_l = OpinionSite._download(self)
        if self.test_mode_enabled():
            return [html_l]
        html_trees = []
        path = "//td[contains(./text(),'Opinion') or contains(./text(), 'PER CURIAM')]/preceding-sibling::td[1]//@href"
        for url in html_l.xpath(path)[:2]:
            html_tree = self._get_html_tree_by_url(url, request_dict)
            html_trees.append(html_tree)
        return html_trees

    def _get_case_names(self):
        case_names = []
        for html_tree in self.html:
            case_names.extend(self._add_case_names(html_tree))
        return case_names

    def _get_download_urls(self):
        download_urls = []
        for html_tree in self.html:
            download_urls.extend(self._add_download_urls(html_tree))
        return download_urls

    def _get_case_dates(self):
        case_dates = []
        for html_tree in self.html:
            case_dates.extend(self._add_dates(html_tree))
        return case_dates

    def _get_precedential_statuses(self):
        return ['Published'] * len(self.case_dates)

    def _get_docket_numbers(self):
        docket_numbers = []
        for html_tree in self.html:
            docket_numbers.extend(self._add_docket_numbers(html_tree))
        return docket_numbers

    def _get_judges(self):
        judges = []
        for html_tree in self.html:
            judges.extend(self._add_judges(html_tree))
        return judges

    def _add_judges(self, html_tree):
        judges = []
        for e in html_tree.xpath("//strong[contains(., 'J.') or contains(., 'CURIAM')]"):
            text = html.tostring(e, method='text', encoding='unicode')
            text = ' '.join(text.split())
            match = re.search(r'(\w+\s*\w+,?\s?J?\.?)', text)
            if match:
                judge = match.group(1)
                judge = re.sub('^BY ', '', judge)
                if 'CURIAM' in judge:
                    preceding_judge = 'CURIAM'
                else:
                    preceding_judge = judge
                judge_path_format = "%s[preceding::strong[1][contains(., '%s') or text()=':']]//a"
                judge_path = judge_path_format % (self.path_case_paragraph, preceding_judge)
                judge_count = len(html_tree.xpath(judge_path))
                judges.extend([judge] * judge_count)
        nr_of_links = self._get_link_count_from_html(html_tree)

        nr_of_judges = len(judges)
        if nr_of_judges < nr_of_links:
            for i in range(nr_of_links - nr_of_judges):
                judges.insert(0, '')
        return judges

    def _add_dates(self, html_tree):
        dates = []
        text = ' '.join(list(html_tree.xpath("//text()[contains(normalize-space(), 'day of')]")))
        text = ' '.join(text.split())
        if text:
            match = re.search(r'\D*((\d{1,2})\w{2} day of (\w+),\s*(\d{4})).*', text)
            date_text = '{day}{month}{year}'.format(day=match.group(2), month=match.group(3), year=match.group(4))
            case_date = date.fromtimestamp(time.mktime(time.strptime(date_text, '%d%B%Y')))
            dates.extend([case_date] * self._get_link_count_from_html(html_tree))
        return dates

    def _add_download_urls(self, html_tree):
        path = "%s/@href" % self.path_case_link
        return html_tree.xpath(path)

    def _add_case_names(self, html_tree):
        case_names = []
        for element in self._get_links_from_html(html_tree):
            text = ' '.join(list(element.xpath(".//text()")))
            text = ' '.join(text.split())
            if text:
                try:
                    case_name = re.search(r'(\d+ ?-\w{1,2} ?- ?\d+)(?!.*\d+ ?-\w{1,2} ?- ?\d+)\s*(.*)', text).group(2)
                    if case_name:
                        pass
                    else:
                        name_element = element.xpath("./parent::p[1]/text()")
                        text = ' '.join(name_element.split())
                        case_name = re.search(r'(\w+.+)+', text).group(1)
                except AttributeError:
                    case_name = re.search(r'(\w+.+)+', text).group(1)
                case_names.append(titlecase(case_name))

        return case_names

    def _add_docket_numbers(self, html_tree):
        regex = r'(\d+ ?-.*- ?\d+)\s*(.*)'
        docket_numbers = []
        for element in self._get_links_from_html(html_tree):
            text = ' '.join(list(element.xpath(".//text()")))
            text = ' '.join(text.split())
            if text:
                try:
                    docket_numbers.append(re.search(regex, text).group(1))
                except AttributeError:
                    name_element = element.xpath("./parent::p[1]//text()")
                    text2 = ' '.join(list(name_element))
                    text2 = ' '.join(text2.split())
                    if text2:
                        docket_numbers.append(re.search(regex, text2).group(1))
        return docket_numbers

    def _get_link_count_from_html(self, html_tree):
        return len(self._get_links_from_html(html_tree))

    def _get_links_from_html(self, html_tree):
        return html_tree.xpath(self.path_case_link)
